import os
import shutil
import textwrap

import pytest

from lektor.builder import Builder
from lektor.db import Database
from lektor.environment import Environment
from lektor.project import Project

sep = os.path.sep


@pytest.fixture(scope="function")
def theme_project_tmpdir(tmpdir):
    """Copy themes-project to a temp dir, and copy demo-project content to it."""
    themes_dir = os.path.join(os.path.dirname(__file__), "themes-project")
    content_dir = os.path.join(os.path.dirname(__file__), "demo-project", "content")

    temp_dir = tmpdir.mkdir("temp").join("themes-project")

    shutil.copytree(themes_dir, str(temp_dir))
    shutil.copytree(content_dir, str(temp_dir.join("content")))

    return temp_dir


@pytest.fixture(scope="function")
def theme_project(theme_project_tmpdir, request):
    """Return the theme project created in a temp dir.

    Could be parametrize, if request.param=False themes variables won't be set
    """
    try:
        with_themes_var = request.param
    except AttributeError:
        with_themes_var = True

    # Create the .lektorproject file
    lektorfile_text = textwrap.dedent(
        """
        [project]
        name = Themes Project
        {}
    """.format(
            "themes = blog_theme, project_theme" if with_themes_var else ""
        )
    )

    theme_project_tmpdir.join("themes.lektorproject").write_text(
        lektorfile_text, "utf8", ensure=True
    )
    return Project.from_path(str(theme_project_tmpdir))


@pytest.fixture(scope="function")
def theme_env(theme_project):

    return Environment(theme_project)


@pytest.fixture(scope="function")
def theme_pad(theme_env):

    return Database(theme_env).new_pad()


@pytest.fixture(scope="function")
def theme_builder(theme_pad, tmpdir):

    return Builder(theme_pad, str(tmpdir.mkdir("output")))


@pytest.mark.parametrize(
    "theme_project, themes",
    [
        (True, ["blog_theme", "project_theme"]),
        # when themes variables isn't set themes are only loaded in the env
        (False, []),
    ],
    indirect=["theme_project"],
)
def test_loading_theme_variable(theme_project, themes):
    assert theme_project.themes == themes


@pytest.mark.parametrize("theme_project", (True,), indirect=True)
def test_loading_theme_path(theme_env):
    assert [os.path.basename(p) for p in theme_env.theme_paths] == [
        "blog_theme",
        "project_theme",
    ]


@pytest.mark.parametrize("theme_project", (False,), indirect=True)
def test_loading_theme_path_variable_dont_set(theme_env):
    """Themes will be loaded, but the order could change."""
    paths = [os.path.basename(p) for p in theme_env.theme_paths]
    assert len(paths) == 2
    for path in paths:
        assert path in ["blog_theme", "project_theme"]


@pytest.mark.parametrize(
    "asset_name, found_in",
    [
        # - themes-project/assets/asset.txt
        # only exist in themes-project assets will be loaded from there
        ("asset.txt", "root"),
        # - themes-project/assets/dummy.txt
        # - themes-project/themes/blog_theme/assets/dummy.txt
        # wil be loaded from themes-project assets not from blog_theme assets
        ("dummy.txt", "root"),
        # - themes-project/themes/blog_theme/static/blog.css
        # only exist in blog_theme assets will be loaded from there
        ("static/blog.css", "blog"),
        # - themes-project/themes/project_theme/static/project.css
        # only exist in project_theme assets will be loaded from there
        ("static/project.css", "project"),
        # - themes-project/themes/blog_theme/assets/dummy2.txt
        # - themes-project/themes/project_theme/assets/dummy2.txt
        # wil be loaded from blog_theme assets because is included first
        ("dummy2.txt", "blog"),
    ],
)
def test_theme_asset_loading(theme_pad, asset_name, found_in):
    """Test loading assets from theme project.

    Loading should take in account the order of the themes
    """
    path_list = theme_pad.get_asset(asset_name).source_filename.split(sep)
    path_list_from_url = theme_pad.resolve_url_path(asset_name).source_filename.split(
        sep
    )

    assert path_list == path_list_from_url

    assert (found_in == "root") == ("themes" not in path_list)
    assert (found_in == "blog") == ("blog_theme" in path_list)
    assert (found_in == "project") == ("project_theme" in path_list)


@pytest.mark.parametrize(
    "url, datamodel_name, found_in",
    [
        # - themes-project/themes/blog_theme/models/blog.ini
        # only exist in blog_theme will be loaded from there
        ("/blog", "Blog", "blog"),
        # - themes-project/themes/project_theme/models/projects.ini
        # only exist in project_theme will be loaded from there
        ("/projects", "Projects", "project"),
        # - themes-project/models/blog-post.ini
        # - themes-project/themes/blog_theme/models/blog-post.ini
        # will be loaded from themes-project models
        ("/blog/post1", "Blog Post", "root"),
        # - themes-project/models/page.ini
        # only exist in themes-project will be loaded from there
        ("/", "Page", "root"),
        # - themes-project/themes/blog_theme/models/project.ini
        # - themes-project/themes/project_theme/models/project.ini
        # wil be loaded from blog_theme assets because is included first
        ("/projects/bagpipe", "Project", "blog"),
    ],
)
def test_theme_models_loading(theme_pad, url, datamodel_name, found_in):
    """Test loading models from theme project.

    Loading should take in account the order of the themes
    """
    assert datamodel_name == theme_pad.get(url).datamodel.name

    path_list = theme_pad.get(url).datamodel.filename.split(sep)

    assert (found_in == "root") == ("themes" not in path_list)
    assert (found_in == "blog") == ("blog_theme" in path_list)
    assert (found_in == "project") == ("project_theme" in path_list)


@pytest.mark.parametrize(
    "template_name, found_in",
    [
        # - themes-project/templates/layout.html
        # - themes-project/themes/blog_theme/templates/layout.html
        # - themes-project/themes/project_theme/templates/layout.html
        # will be loaded from themes-project templates
        ("layout.html", "root"),
        # - themes-project/themes/blog_theme/templates/blog.html
        # only exist in blog_theme will be loaded from there
        ("blog.html", "blog"),
        # - themes-project/themes/project_theme/templates/project.html
        # only exist in project_theme will be loaded from there
        ("project.html", "project"),
        # - themes-project/templates/page.html
        # only exist in themes-project will be loaded from there
        ("page.html", "root"),
        # - themes-project/themes/blog_theme/templates/dummy.html
        # - themes-project/themes/project_theme/templates/dummy.html
        # wil be loaded from blog_theme assets because is included first
        ("dummy.html", "blog"),
    ],
)
def test_theme_templates_loading(theme_env, template_name, found_in):
    """Test loading templates from theme project.

    Loading should take in account the order of the themes
    """
    path_list = theme_env.jinja_env.get_template(template_name).filename.split(sep)

    assert (found_in == "root") == ("themes" not in path_list)
    assert (found_in == "blog") == ("blog_theme" in path_list)
    assert (found_in == "project") == ("project_theme" in path_list)



from markers import imagemagick


def get_child_sources(prog):
    return sorted(list(prog.iter_child_sources()), key=lambda x: x["_id"])


@imagemagick
def test_basic_build(pad, builder):
    root = pad.root

    prog, build_state = builder.build(root)
    assert prog.source is root
    assert build_state.failed_artifacts == []

    (artifact,) = prog.artifacts
    # Root and its thumbnail image were updated.
    assert artifact in build_state.updated_artifacts
    assert artifact.artifact_name == "index.html"
    assert set(artifact.sources) == set(root.iter_source_filenames())
    assert artifact.updated
    assert artifact.extra is None
    assert artifact.config_hash is None


def test_child_sources_basic(pad, builder):
    extra = pad.get("/extra")

    prog, _ = builder.build(extra)
    child_sources = get_child_sources(prog)

    assert [x["_id"] for x in child_sources] == [
        "a",
        "b",
        "file.ext",
        "hello.txt",
        "paginated",
        "slash-slug",
    ]


def test_child_sources_pagination(pad, builder):
    projects = pad.get("/projects")

    def ids(sources):
        return [x["_id"] for x in sources]

    prog, _ = builder.build(projects)

    child_sources = get_child_sources(prog)
    assert ids(child_sources) == [
        "attachment.txt",
        "filtered",
        "projects",
        "projects",
        "secret",
    ]

    page1 = child_sources[2]
    assert page1["_id"] == "projects"
    assert page1.page_num == 1
    page2 = child_sources[3]
    assert page2["_id"] == "projects"
    assert page2.page_num == 2

    prog, _ = builder.build(page1)
    child_sources_p1 = get_child_sources(prog)

    assert ids(child_sources_p1) == [
        "bagpipe",
        "coffee",
        "master",
        "oven",
    ]

    prog, _ = builder.build(page2)
    child_sources_p2 = get_child_sources(prog)

    assert ids(child_sources_p2) == [
        "postage",
        "slave",
        "wolf",
    ]


def test_basic_artifact_current_test(pad, builder, reporter):
    post1 = pad.get("blog/post1")

    def build():
        reporter.clear()
        prog, _ = builder.build(post1)
        return prog.artifacts[0]

    artifact = build()

    assert reporter.get_major_events() == [
        (
            "enter-source",
            {
                "source": post1,
            },
        ),
        (
            "start-artifact-build",
            {
                "artifact": artifact,
                "is_current": False,
            },
        ),
        (
            "build-func",
            {
                "func": "lektor.build_programs.PageBuildProgram",
            },
        ),
        (
            "finish-artifact-build",
            {
                "artifact": artifact,
            },
        ),
        (
            "leave-source",
            {
                "source": post1,
            },
        ),
    ]

    assert set(reporter.get_recorded_dependencies()) == {
        "Website.lektorproject",
        "content/blog/post1/contents+en.lr",
        "content/blog/post1/contents.lr",
        "templates/blog-post.html",
        "templates/layout.html",
        "models/blog-post.ini",
    }

    assert artifact.is_current

    artifact = build()

    assert artifact.is_current

    assert reporter.get_major_events() == [
        (
            "enter-source",
            {
                "source": post1,
            },
        ),
        (
            "start-artifact-build",
            {
                "artifact": artifact,
                "is_current": True,
            },
        ),
        (
            "build-func",
            {
                "func": "lektor.build_programs.PageBuildProgram",
            },
        ),
        (
            "finish-artifact-build",
            {
                "artifact": artifact,
            },
        ),
        (
            "leave-source",
            {
                "source": post1,
            },
        ),
    ]


def test_basic_template_rendering(pad, builder):
    root = pad.root

    prog, _ = builder.build(root)
    artifact = prog.artifacts[0]

    with artifact.open("rb") as f:
        rv = f.read().decode("utf-8")

    assert artifact.artifact_name == "index.html"

    assert "<title>My Website</title>" in rv
    assert "<h1>Welcome</h1>" in rv
    assert '<link href="static/style.css" rel="stylesheet">' in rv
    assert "<p>Welcome to this pretty nifty website.</p>" in rv


def test_attachment_copying(pad, builder):
    root = pad.root
    text_file = root.attachments.get("hello.txt")

    prog, _ = builder.build(text_file)
    artifact = prog.artifacts[0]

    assert artifact.artifact_name == "hello.txt"

    with artifact.open("rb") as f:
        rv = f.read().decode("utf-8").strip()
        assert rv == "Hello I am an Attachment"


def test_asset_processing(pad, builder):
    static = pad.asset_root.get_child("static")

    prog, _ = builder.build(static)
    assets = list(prog.iter_child_sources())
    assert len(assets) == 1
    assert assets[0].name == "demo.css"

    prog, _ = builder.build(assets[0])
    with prog.artifacts[0].open("rb") as f:
        rv = f.read().decode("utf-8").strip()
        assert "color: red" in rv


def test_included_assets(pad, builder):
    # In demo-project/Website.lektorproject, included_assets = "_*".
    root = pad.asset_root

    prog, _ = builder.build(root)
    assets = list(prog.iter_child_sources())
    assert "_include_me_despite_underscore" in [a.name for a in assets]


def test_excluded_assets(pad, builder):
    # In demo-project/Website.lektorproject, excluded_assets = "foo*".
    root = pad.asset_root

    prog, _ = builder.build(root)
    assets = list(prog.iter_child_sources())
    assert "foo-prefix-makes-me-excluded" not in [a.name for a in assets]


def test_iter_child_pages(child_sources_test_project_builder):
    # Test that child sources are built even if they're filtered out by a
    # pagination query like "this.children.filter(F._model == 'doesnt-exist')".
    builder = child_sources_test_project_builder
    pad = builder.pad
    prog, _ = builder.build(pad.root)
    assert builder.pad.get("filtered-page") in prog.iter_child_sources()


def test_iter_child_attachments(child_sources_test_project_builder):
    # Test that attachments are built, even if a pagination has no items.
    builder = child_sources_test_project_builder
    pad = builder.pad
    prog, _ = builder.build(pad.root)
    assert builder.pad.get("attachment.txt") in prog.iter_child_sources()


def test_record_is_file(pad, builder):
    record = pad.get("/extra/file.ext")

    prog, _ = builder.build(record)
    (artifact,) = prog.artifacts
    assert artifact.artifact_name == "extra/file.ext"


def test_slug_contains_slash(pad, builder):
    record = pad.get("/extra/slash-slug")

    prog, _ = builder.build(record)
    (artifact,) = prog.artifacts
    assert artifact.artifact_name == "extra/long/path/index.html"



def _line_is_dashes(line):
    line = line.strip()
    return line == "-" * len(line) and len(line) >= 3


def _process_buf(buf):
    for idx, line in enumerate(buf):
        if _line_is_dashes(line):
            line = line[1:]
        buf[idx] = line

    if buf and buf[-1][-1:] == "\n":
        buf[-1] = buf[-1][:-1]

    return buf[:]


def tokenize(iterable, interesting_keys=None, encoding=None):
    """This tokenizes an iterable of newlines as bytes into key value
    pairs out of the lektor bulk format.  By default it will process all
    fields, but optionally it can skip values of uninteresting keys and
    will instead yield `None`.  The values are left as list of decoded
    lines with their endings preserved.

    This will not perform any other processing on the data other than
    decoding and basic tokenizing.
    """
    key = []
    buf = []
    want_newline = False
    is_interesting = True

    def _flush_item():
        the_key = key[0]
        if not is_interesting:
            value = None
        else:
            value = _process_buf(buf)
        del key[:], buf[:]
        return the_key, value

    if encoding is not None:
        iterable = (x.decode(encoding, "replace") for x in iterable)

    for line in iterable:
        line = line.rstrip("\r\n") + "\n"

        if line.rstrip() == "---":
            want_newline = False
            if key:
                yield _flush_item()
        elif key:
            if want_newline:
                want_newline = False
                if not line.strip():
                    continue
            if is_interesting:
                buf.append(line)
        else:
            bits = line.split(":", 1)
            if len(bits) == 2:
                key = [bits[0].strip()]
                if interesting_keys is None:
                    is_interesting = True
                else:
                    is_interesting = key[0] in interesting_keys
                if is_interesting:
                    first_bit = bits[1].strip("\t ")
                    if first_bit.strip():
                        buf = [first_bit]
                    else:
                        buf = []
                        want_newline = True

    if key:
        yield _flush_item()


def serialize(iterable, encoding=None):
    """Serializes an iterable of key value pairs into a stream of
    string chunks.  If an encoding is provided, it will be encoded into that.

    This is primarily used by the editor to write back data to a source file.
    """

    def _produce(item, escape=False):
        if escape:
            if _line_is_dashes(item):
                item = "-" + item
        if encoding is not None:
            item = item.encode(encoding)
        return item

    for idx, (key, value) in enumerate(iterable):
        value = value.replace("\r\n", "\n").replace("\r", "\n")
        if idx > 0:
            yield _produce("---\n")
        if "\n" in value or value.strip("\t ") != value:
            yield _produce(key + ":\n")
            yield _produce("\n")
            for line in value.splitlines(True):
                yield _produce(line, escape=True)
            yield _produce("\n")
        else:
            yield _produce("%s: %s\n" % (key, value))



import pytest

from lektor.imagetools import find_imagemagick


try:
    im_path = find_imagemagick()
except RuntimeError:
    im_path = None


imagemagick = pytest.mark.skipif(
    not im_path,
    reason="imagemagick required but not found",
)



# -*- coding: utf-8 -*-
# pylint: disable=too-many-lines
import errno
import functools
import hashlib
import operator
import os
import posixpath
import warnings
from collections import OrderedDict
from datetime import timedelta
from itertools import islice
from operator import methodcaller

from jinja2 import is_undefined
from jinja2 import Undefined
from jinja2.exceptions import UndefinedError
from jinja2.utils import LRUCache
from werkzeug.urls import url_join
from werkzeug.utils import cached_property

from lektor import metaformat
from lektor.assets import Directory
from lektor.constants import PRIMARY_ALT
from lektor.context import Context
from lektor.context import get_ctx
from lektor.databags import Databags
from lektor.datamodel import load_datamodels
from lektor.datamodel import load_flowblocks
from lektor.editor import make_editor_session
from lektor.filecontents import FileContents
from lektor.imagetools import get_image_info
from lektor.imagetools import make_image_thumbnail
from lektor.imagetools import read_exif
from lektor.imagetools import ThumbnailMode
from lektor.sourceobj import SourceObject
from lektor.sourceobj import VirtualSourceObject
from lektor.utils import cleanup_path
from lektor.utils import fs_enc
from lektor.utils import locate_executable
from lektor.utils import make_relative_url
from lektor.utils import sort_normalize_string
from lektor.utils import split_virtual_path
from lektor.utils import untrusted_to_os_path
from lektor.videotools import get_video_info
from lektor.videotools import make_video_thumbnail

# pylint: disable=no-member


def get_alts(source=None, fallback=False):
    """Given a source this returns the list of all alts that the source
    exists as.  It does not include fallbacks unless `fallback` is passed.
    If no source is provided all configured alts are returned.  If alts are
    not configured at all, the return value is an empty list.
    """
    if source is None:
        ctx = get_ctx()
        if ctx is None:
            raise RuntimeError("This function requires the context to be supplied.")
        pad = ctx.pad
    else:
        pad = source.pad
    alts = list(pad.config.iter_alternatives())
    if alts == [PRIMARY_ALT]:
        return []

    rv = alts

    # If a source is provided and it's not virtual, we look up all alts
    # of the path on the pad to figure out which records exist.
    if source is not None and "@" not in source.path:
        rv = []
        for alt in alts:
            if pad.alt_exists(source.path, alt=alt, fallback=fallback):
                rv.append(alt)

    return rv


def _require_ctx(record):
    ctx = get_ctx()
    if ctx is None:
        raise RuntimeError(
            "This operation requires a context but none was " "on the stack."
        )
    if ctx.pad is not record.pad:
        raise RuntimeError(
            "The context on the stack does not match the " "pad of the record."
        )
    return ctx


def _is_content_file(filename, alt=PRIMARY_ALT):
    if filename == "contents.lr":
        return True
    if alt != PRIMARY_ALT and filename == "contents+%s.lr" % alt:
        return True
    return False


class _CmpHelper:
    def __init__(self, value, reverse):
        self.value = value
        self.reverse = reverse

    @staticmethod
    def coerce(a, b):
        if isinstance(a, str) and isinstance(b, str):
            return sort_normalize_string(a), sort_normalize_string(b)
        if type(a) is type(b):
            return a, b
        if isinstance(a, Undefined) or isinstance(b, Undefined):
            if isinstance(a, Undefined):
                a = None
            if isinstance(b, Undefined):
                b = None
            return a, b
        if isinstance(a, (int, float)):
            try:
                return a, type(a)(b)
            except (ValueError, TypeError, OverflowError):
                pass
        if isinstance(b, (int, float)):
            try:
                return type(b)(a), b
            except (ValueError, TypeError, OverflowError):
                pass
        return a, b

    def __eq__(self, other):
        a, b = self.coerce(self.value, other.value)
        return a == b

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        a, b = self.coerce(self.value, other.value)
        try:
            if self.reverse:
                return b < a
            return a < b
        except TypeError:
            # Put None at the beginning if reversed, else at the end.
            if self.reverse:
                return a is not None
            return a is None

    def __gt__(self, other):
        return not (self.__lt__(other) or self.__eq__(other))

    def __le__(self, other):
        return self.__lt__(other) or self.__eq__(other)

    def __ge__(self, other):
        return not self.__lt__(other)


def _auto_wrap_expr(value):
    if isinstance(value, Expression):
        return value
    return _Literal(value)


def save_eval(filter, record):
    try:
        return filter.__eval__(record)
    except UndefinedError as e:
        return Undefined(e.message)


class Expression:
    def __eval__(self, record):
        # pylint: disable=no-self-use
        return record

    def __eq__(self, other):
        return _BinExpr(self, _auto_wrap_expr(other), operator.eq)

    def __ne__(self, other):
        return _BinExpr(self, _auto_wrap_expr(other), operator.ne)

    def __and__(self, other):
        return _BinExpr(self, _auto_wrap_expr(other), operator.and_)

    def __or__(self, other):
        return _BinExpr(self, _auto_wrap_expr(other), operator.or_)

    def __gt__(self, other):
        return _BinExpr(self, _auto_wrap_expr(other), operator.gt)

    def __ge__(self, other):
        return _BinExpr(self, _auto_wrap_expr(other), operator.ge)

    def __lt__(self, other):
        return _BinExpr(self, _auto_wrap_expr(other), operator.lt)

    def __le__(self, other):
        return _BinExpr(self, _auto_wrap_expr(other), operator.le)

    def contains(self, item):
        return _ContainmentExpr(self, _auto_wrap_expr(item))

    def startswith(self, other):
        return _BinExpr(
            self,
            _auto_wrap_expr(other),
            lambda a, b: str(a).lower().startswith(str(b).lower()),
        )

    def endswith(self, other):
        return _BinExpr(
            self,
            _auto_wrap_expr(other),
            lambda a, b: str(a).lower().endswith(str(b).lower()),
        )

    def startswith_cs(self, other):
        return _BinExpr(
            self,
            _auto_wrap_expr(other),
            lambda a, b: str(a).startswith(str(b)),
        )

    def endswith_cs(self, other):
        return _BinExpr(
            self,
            _auto_wrap_expr(other),
            lambda a, b: str(a).endswith(str(b)),
        )

    def false(self):
        return _IsBoolExpr(self, False)

    def true(self):
        return _IsBoolExpr(self, True)


# Query helpers for the template engine
setattr(Expression, "and", lambda x, o: x & o)
setattr(Expression, "or", lambda x, o: x | o)


class _CallbackExpr(Expression):
    def __init__(self, func):
        self.func = func

    def __eval__(self, record):
        return self.func(record)


class _IsBoolExpr(Expression):
    def __init__(self, expr, true):
        self.__expr = expr
        self.__true = true

    def __eval__(self, record):
        val = self.__expr.__eval__(record)
        return (
            not is_undefined(val) and val not in (None, 0, False, "")
        ) == self.__true


class _Literal(Expression):
    def __init__(self, value):
        self.__value = value

    def __eval__(self, record):
        return self.__value


class _BinExpr(Expression):
    def __init__(self, left, right, op):
        self.__left = left
        self.__right = right
        self.__op = op

    def __eval__(self, record):
        return self.__op(self.__left.__eval__(record), self.__right.__eval__(record))


class _ContainmentExpr(Expression):
    def __init__(self, seq, item):
        self.__seq = seq
        self.__item = item

    def __eval__(self, record):
        seq = self.__seq.__eval__(record)
        item = self.__item.__eval__(record)
        if isinstance(item, Record):
            item = item["_id"]
        return item in seq


class _RecordQueryField(Expression):
    def __init__(self, field):
        self.__field = field

    def __eval__(self, record):
        try:
            return record[self.__field]
        except KeyError:
            return Undefined(obj=record, name=self.__field)


class _RecordQueryProxy:
    def __getattr__(self, name):
        if name[:2] != "__":
            return _RecordQueryField(name)
        raise AttributeError(name)

    def __getitem__(self, name):
        try:
            return self.__getattr__(name)
        except AttributeError as error:
            raise KeyError(name) from error


F = _RecordQueryProxy()


class Record(SourceObject):
    source_classification = "record"
    supports_pagination = False

    def __init__(self, pad, data, page_num=None):
        SourceObject.__init__(self, pad)
        self._data = data
        self._bound_data = {}
        if page_num is not None and not self.supports_pagination:
            raise RuntimeError(
                "%s does not support pagination" % self.__class__.__name__
            )
        self.page_num = page_num

    @property
    def record(self):
        return self

    @property
    def datamodel(self):
        """Returns the data model for this record."""
        try:
            return self.pad.db.datamodels[self._data["_model"]]
        except LookupError:
            # If we cannot find the model we fall back to the default one.
            return self.pad.db.default_model

    @property
    def alt(self):
        """Returns the alt of this source object."""
        return self["_alt"]

    @property
    def is_hidden(self):
        """Indicates if a record is hidden.  A record is considered hidden
        if the record itself is hidden or the parent is.
        """
        if not is_undefined(self._data["_hidden"]):
            return self._data["_hidden"]
        return self._is_considered_hidden()

    def _is_considered_hidden(self):
        parent = self.parent
        if parent is None:
            return False

        hidden_children = parent.datamodel.child_config.hidden
        if hidden_children is not None:
            return hidden_children
        return parent.is_hidden

    @property
    def is_discoverable(self):
        """Indicates if the page is discoverable without knowing the URL."""
        return self._data["_discoverable"] and not self.is_hidden

    @cached_property
    def pagination(self):
        """Returns the pagination controller for the record."""
        if not self.supports_pagination:
            raise AttributeError()
        return self.datamodel.pagination_config.get_pagination_controller(self)

    @cached_property
    def contents(self):
        return FileContents(self.source_filename)

    def get_fallback_record_label(self, lang):
        if not self["_id"]:
            return "(Index)"
        return self["_id"].replace("-", " ").replace("_", " ").title()

    def get_record_label_i18n(self):
        rv = {}
        for lang, _ in (self.datamodel.label_i18n or {}).items():
            label = self.datamodel.format_record_label(self, lang)
            if not label:
                label = self.get_fallback_record_label(lang)
            rv[lang] = label
        # Fill in english if missing
        if "en" not in rv:
            rv["en"] = self.get_fallback_record_label("en")
        return rv

    @property
    def record_label(self):
        return (self.get_record_label_i18n() or {}).get("en")

    @property
    def url_path(self):
        # This is redundant (it's the same as the inherited
        # SourceObject.url_path) but is here to silence
        # pylint ("W0223: Method 'url_path' is abstract in class
        # 'SourceObject' but is not overridden (abstract-method)"),
        # as well as to document that Record is an abstract class.
        raise NotImplementedError()

    def _get_clean_url_path(self):
        """The "clean" URL path, before modification to account for alt and
        page_num and without any leading '/'
        """
        bits = [self["_slug"]]
        parent = self.parent
        while parent is not None:
            slug = parent["_slug"]
            head, sep, tail = slug.rpartition("/")
            if "." in tail:
                # https://www.getlektor.com/docs/content/urls/#content-below-dotted-slugs
                slug = head + sep + f"_{tail}"
            bits.append(slug)
            parent = parent.parent
        return "/".join(reversed(bits)).strip("/")

    def _get_url_path(self, alt=None):
        """The target path where the record should end up, after adding prefix/suffix
        for the specified alt (but before accounting for any page_num).

        Note that some types of records (Attachments), are only
        created for the primary alternative.
        """
        clean_path = self._get_clean_url_path()
        config = self.pad.config
        if config.primary_alternative:
            # alternatives are configured
            if alt is None:
                alt = config.primary_alternative
            prefix, suffix = config.get_alternative_url_span(alt)
            # XXX: 404.html with suffix -de becomes 404.html-de but should
            # actually become 404-de.html
            clean_path = prefix.lstrip("/") + clean_path + suffix.rstrip("/")
        return "/" + clean_path.rstrip("/")

    @property
    def path(self):
        return self["_path"]

    def get_sort_key(self, fields):
        """Returns a sort key for the given field specifications specific
        for the data in the record.
        """
        rv = [None] * len(fields)
        for idx, field in enumerate(fields):
            if field[:1] == "-":
                field = field[1:]
                reverse = True
            else:
                field = field.lstrip("+")
                reverse = False
            try:
                value = self[field]
            except KeyError:
                value = None
            rv[idx] = _CmpHelper(value, reverse)
        return rv

    def __contains__(self, name):
        return name in self._data and not is_undefined(self._data[name])

    def __getitem__(self, name):
        rv = self._bound_data.get(name, Ellipsis)
        if rv is not Ellipsis:
            return rv
        rv = self._data[name]
        if hasattr(rv, "__get__"):
            rv = rv.__get__(self)
        self._bound_data[name] = rv
        return rv

    def __eq__(self, other):
        if self is other:
            return True
        if self.__class__ != other.__class__:
            return False
        return self["_path"] == other["_path"]

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.path)

    def __repr__(self):
        return "<%s model=%r path=%r%s%s>" % (
            self.__class__.__name__,
            self["_model"],
            self["_path"],
            self.alt != PRIMARY_ALT and " alt=%r" % self.alt or "",
            self.page_num is not None and " page_num=%r" % self.page_num or "",
        )


class Siblings(VirtualSourceObject):  # pylint: disable=abstract-method
    def __init__(self, record, prev_page, next_page):
        """Virtual source representing previous and next sibling of 'record'."""
        VirtualSourceObject.__init__(self, record)
        self._path = record.path + "@siblings"
        self._prev_page = prev_page
        self._next_page = next_page

    @property
    def path(self):
        # Used as a key in Context.referenced_virtual_dependencies.
        return self._path

    @property
    def prev_page(self):
        return self._prev_page

    @property
    def next_page(self):
        return self._next_page

    def iter_source_filenames(self):
        for page in self._prev_page, self._next_page:
            if page:
                yield page.source_filename

    def _file_infos(self, path_cache):
        for page in self._prev_page, self._next_page:
            if page:
                yield path_cache.get_file_info(page.source_filename)

    def get_mtime(self, path_cache):
        mtimes = [i.mtime for i in self._file_infos(path_cache)]
        return max(mtimes) if mtimes else None

    def get_checksum(self, path_cache):
        sums = "|".join(i.filename_and_checksum for i in self._file_infos(path_cache))

        return sums or None


def siblings_resolver(node, url_path):
    return node.get_siblings()


class Page(Record):
    """This represents a loaded record."""

    is_attachment = False
    supports_pagination = True

    @cached_property
    def path(self):
        rv = self["_path"]
        if self.page_num is not None:
            rv = "%s@%s" % (rv, self.page_num)
        return rv

    @cached_property
    def record(self):
        if self.page_num is None:
            return self
        return self.pad.get(
            self["_path"], persist=self.pad.cache.is_persistent(self), alt=self.alt
        )

    @property
    def source_filename(self):
        if self.alt != PRIMARY_ALT:
            return os.path.join(
                self.pad.db.to_fs_path(self["_path"]), "contents+%s.lr" % self.alt
            )
        return os.path.join(self.pad.db.to_fs_path(self["_path"]), "contents.lr")

    def iter_source_filenames(self):
        yield self.source_filename
        if self.alt != PRIMARY_ALT:
            yield os.path.join(self.pad.db.to_fs_path(self["_path"]), "contents.lr")

    @property
    def url_path(self):
        pg = self.datamodel.pagination_config
        path = self._get_url_path(self.alt)
        _, _, last_part = path.rpartition("/")
        if not pg.enabled:
            if "." in last_part:
                return path
            return path.rstrip("/") + "/"
        if "." in last_part:
            raise RuntimeError(
                "When file extension is provided pagination cannot be used."
            )
        # pagination is enabled
        if self.page_num in (1, None):
            return path.rstrip("/") + "/"
        return f"{path.rstrip('/')}/{pg.url_suffix.strip('/')}/{self.page_num:d}/"

    def resolve_url_path(self, url_path):
        pg = self.datamodel.pagination_config

        # If we hit the end of the url path, then we found our target.
        # However if pagination is enabled we want to resolve the first
        # page instead of the unpaginated version.
        if not url_path:
            if pg.enabled and self.page_num is None:
                return pg.get_record_for_page(self, 1)
            return self

        # Try to resolve the correctly paginated version here.
        if pg.enabled:
            rv = pg.match_pagination(self, url_path)
            if rv is not None:
                return rv

        # When we resolve URLs we also want to be able to explicitly
        # target undiscoverable pages.  Those who know the URL are
        # rewarded.
        q = self.children.include_undiscoverable(True)

        for idx in range(len(url_path)):
            piece = "/".join(url_path[: idx + 1])
            child = q.filter(F._slug == piece).first()
            if child is None:
                attachment = self.attachments.filter(F._slug == piece).first()
                if attachment is None:
                    obj = self.pad.db.env.resolve_custom_url_path(self, url_path)
                    if obj is None:
                        continue
                    node = obj
                else:
                    node = attachment
            else:
                node = child

            rv = node.resolve_url_path(url_path[idx + 1 :])
            if rv is not None:
                return rv
        return None

    @cached_property
    def parent(self):
        """The parent of the record."""
        this_path = self._data["_path"]
        parent_path = posixpath.dirname(this_path)
        if parent_path != this_path:
            return self.pad.get(
                parent_path, persist=self.pad.cache.is_persistent(self), alt=self.alt
            )
        return None

    @property
    def children(self):
        """A query over all children that are not hidden or undiscoverable.
        want undiscoverable then use ``children.include_undiscoverable(True)``.
        """
        repl_query = self.datamodel.get_child_replacements(self)
        if repl_query is not None:
            return repl_query.include_undiscoverable(False)
        return Query(path=self["_path"], pad=self.pad, alt=self.alt)

    @property
    def attachments(self):
        """Returns a query for the attachments of this record."""
        return AttachmentsQuery(path=self["_path"], pad=self.pad, alt=self.alt)

    def has_prev(self):
        return self.get_siblings().prev_page is not None

    def has_next(self):
        return self.get_siblings().next_page is not None

    def get_siblings(self):
        """The next and previous children of this page's parent.

        Uses parent's pagination query, if any, else parent's "children" config.
        """
        siblings = Siblings(self, *self._siblings)
        ctx = get_ctx()
        if ctx:
            ctx.pad.db.track_record_dependency(siblings)
        return siblings

    @cached_property
    def _siblings(self):
        parent = self.parent
        pagination_enabled = parent.datamodel.pagination_config.enabled

        # Don't track dependencies for this part.
        with Context(pad=self.pad):
            if pagination_enabled:
                pagination = parent.pagination
                siblings = list(pagination.config.get_pagination_query(parent))
            else:
                siblings = list(parent.children)

            prev_item, next_item = None, None
            try:
                me = siblings.index(self)
            except ValueError:
                # Self not in parents.children or not in parents.pagination.
                pass
            else:
                if me > 0:
                    prev_item = siblings[me - 1]

                if me + 1 < len(siblings):
                    next_item = siblings[me + 1]

        return prev_item, next_item


class Attachment(Record):
    """This represents a loaded attachment."""

    is_attachment = True

    @property
    def source_filename(self):
        if self.alt != PRIMARY_ALT:
            suffix = "+%s.lr" % self.alt
        else:
            suffix = ".lr"
        return self.pad.db.to_fs_path(self["_path"]) + suffix

    def _is_considered_hidden(self):
        # Attachments are only considered hidden if they have been
        # configured as such.  This means that even if a record itself is
        # hidden, the attachments by default will not.
        parent = self.parent
        if parent is None:
            return False
        return parent.datamodel.attachment_config.hidden

    @property
    def record(self):
        return self

    @property
    def attachment_filename(self):
        return self.pad.db.to_fs_path(self["_path"])

    @property
    def parent(self):
        """The associated record for this attachment."""
        return self.pad.get(
            self._data["_attachment_for"], persist=self.pad.cache.is_persistent(self)
        )

    @cached_property
    def contents(self):
        return FileContents(self.attachment_filename)

    def get_fallback_record_label(self, lang):
        return self["_id"]

    def iter_source_filenames(self):
        yield self.source_filename
        yield self.attachment_filename

    @property
    def url_path(self):
        # Attachments are only emitted for the primary alternative.
        primary_alt = self.pad.config.primary_alternative or PRIMARY_ALT
        return self._get_url_path(alt=primary_alt)


class Image(Attachment):
    """Specific class for image attachments."""

    def __init__(self, pad, data, page_num=None):
        Attachment.__init__(self, pad, data, page_num)
        self._image_info = None
        self._exif_cache = None

    def _get_image_info(self):
        if self._image_info is None:
            with open(self.attachment_filename, "rb") as f:
                self._image_info = get_image_info(f)
        return self._image_info

    @property
    def exif(self):
        """Provides access to the exif data."""
        if self._exif_cache is None:
            with open(self.attachment_filename, "rb") as f:
                self._exif_cache = read_exif(f)
        return self._exif_cache

    @property
    def width(self):
        """The width of the image if possible to determine."""
        rv = self._get_image_info()[1]
        if rv is not None:
            return rv
        return Undefined("Width of image could not be determined.")

    @property
    def height(self):
        """The height of the image if possible to determine."""
        rv = self._get_image_info()[2]
        if rv is not None:
            return rv
        return Undefined("Height of image could not be determined.")

    @property
    def format(self):
        """Returns the format of the image."""
        rv = self._get_image_info()[0]
        if rv is not None:
            return rv
        return Undefined("The format of the image could not be determined.")

    def thumbnail(
        self, width=None, height=None, crop=None, mode=None, upscale=None, quality=None
    ):
        """Utility to create thumbnails."""

        # `crop` exists to preserve backward-compatibility, and will be removed.
        if crop is not None and mode is not None:
            raise ValueError("Arguments `crop` and `mode` are mutually exclusive.")

        if crop is not None:
            warnings.warn(
                'The `crop` argument is deprecated. Use `mode="crop"` instead.'
            )
            mode = "crop"

        if mode is None:
            mode = ThumbnailMode.DEFAULT
        else:
            mode = ThumbnailMode(mode)

        if width is not None:
            width = int(width)
        if height is not None:
            height = int(height)

        return make_image_thumbnail(
            _require_ctx(self),
            self.attachment_filename,
            self.url_path,
            width=width,
            height=height,
            mode=mode,
            upscale=upscale,
            quality=quality,
        )


def require_ffmpeg(f):
    """Decorator to help with error messages for ffmpeg template functions."""
    # If both ffmpeg and ffprobe executables are available we don't need to
    # override the function
    if locate_executable("ffmpeg") and locate_executable("ffprobe"):
        return f

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        return Undefined(
            "Unable to locate ffmpeg or ffprobe executable. Is " "it installed?"
        )

    return wrapper


class Video(Attachment):
    """Specific class for video attachments."""

    def __init__(self, pad, data, page_num=None):
        Attachment.__init__(self, pad, data, page_num)
        self._video_info = None

    def _get_video_info(self):
        if self._video_info is None:
            try:
                self._video_info = get_video_info(self.attachment_filename)
            except RuntimeError:
                # A falsy value ensures we don't retry this video again
                self._video_info = False
        return self._video_info

    @property
    @require_ffmpeg
    def width(self):
        """Returns the width of the video if possible to determine."""
        rv = self._get_video_info()
        if rv:
            return rv["width"]
        return Undefined("The width of the video could not be determined.")

    @property
    @require_ffmpeg
    def height(self):
        """Returns the height of the video if possible to determine."""
        rv = self._get_video_info()
        if rv:
            return rv["height"]
        return Undefined("The height of the video could not be determined.")

    @property
    @require_ffmpeg
    def duration(self):
        """Returns the duration of the video if possible to determine."""
        rv = self._get_video_info()
        if rv:
            return rv["duration"]
        return Undefined("The duration of the video could not be determined.")

    @require_ffmpeg
    def frame(self, seek=None):
        """Returns a VideoFrame object that is thumbnailable like an Image."""
        rv = self._get_video_info()
        if not rv:
            return Undefined("Unable to get video properties.")

        if seek is None:
            seek = rv["duration"] / 2
        return VideoFrame(self, seek)


class VideoFrame:
    """Representation of a specific frame in a VideoAttachment.

    This is currently only useful for thumbnails, but in the future it might
    work like an ImageAttachment.
    """

    def __init__(self, video, seek):
        self.video = video

        if not isinstance(seek, timedelta):
            seek = timedelta(seconds=seek)

        if seek < timedelta(0):
            raise ValueError("Seek distance must not be negative")
        if video.duration and seek > video.duration:
            raise ValueError("Seek distance must not be outside the video duration")

        self.seek = seek

    def __str__(self):
        raise NotImplementedError(
            "It is currently not possible to use video "
            "frames directly, use .thumbnail()."
        )

    __unicode__ = __str__

    @require_ffmpeg
    def thumbnail(self, width=None, height=None, mode=None, upscale=None, quality=None):
        """Utility to create thumbnails."""
        if mode is None:
            mode = ThumbnailMode.DEFAULT
        else:
            mode = ThumbnailMode(mode)

        video = self.video
        return make_video_thumbnail(
            _require_ctx(video),
            video.attachment_filename,
            video.url_path,
            seek=self.seek,
            width=width,
            height=height,
            mode=mode,
            upscale=upscale,
            quality=quality,
        )


attachment_classes = {
    "image": Image,
    "video": Video,
}


class Query:
    """Object that helps finding records.  The default configuration
    only finds pages.
    """

    def __init__(self, path, pad, alt=PRIMARY_ALT):
        self.path = path
        self.pad = pad
        self.alt = alt
        self._include_pages = True
        self._include_attachments = False
        self._order_by = None
        self._filters = None
        self._pristine = True
        self._limit = None
        self._offset = None
        self._include_hidden = None
        self._include_undiscoverable = False
        self._page_num = None
        self._filter_func = None

    def __get_lektor_param_hash__(self, h):
        h.update(str(self.alt))
        h.update(str(self._include_pages))
        h.update(str(self._include_attachments))
        h.update("(%s)" % "|".join(self._order_by or ()).encode("utf-8"))
        h.update(str(self._limit))
        h.update(str(self._offset))
        h.update(str(self._include_hidden))
        h.update(str(self._include_undiscoverable))
        h.update(str(self._page_num))

    @property
    def self(self):
        """Returns the object this query starts out from."""
        return self.pad.get(self.path, alt=self.alt)

    def _clone(self, mark_dirty=False):
        """Makes a flat copy but keeps the other data on it shared."""
        rv = object.__new__(self.__class__)
        rv.__dict__.update(self.__dict__)
        if mark_dirty:
            rv._pristine = False
        return rv

    def _get(self, id, persist=True, page_num=Ellipsis):
        """Low level record access."""
        if page_num is Ellipsis:
            page_num = self._page_num
        return self.pad.get(
            "%s/%s" % (self.path, id), persist=persist, alt=self.alt, page_num=page_num
        )

    def _matches(self, record):
        include_hidden = self._include_hidden
        if include_hidden is not None:
            if not self._include_hidden and record.is_hidden:
                return False
        if not self._include_undiscoverable and record.is_undiscoverable:
            return False
        for filter in self._filters or ():
            if not save_eval(filter, record):
                return False
        return True

    def _iterate(self):
        """Low level record iteration."""
        # If we iterate over children we also need to track those
        # dependencies.  There are two ways in which we track them.  The
        # first is through the start record of the query.  If that does
        # not work for whatever reason (because it does not exist for
        # instance).
        self_record = self.pad.get(self.path, alt=self.alt)
        if self_record is not None:
            self.pad.db.track_record_dependency(self_record)

        # We also always want to record the path itself as dependency.
        ctx = get_ctx()
        if ctx is not None:
            ctx.record_dependency(self.pad.db.to_fs_path(self.path))

        for name, _, is_attachment in self.pad.db.iter_items(self.path, alt=self.alt):
            if not (
                (is_attachment == self._include_attachments)
                or (not is_attachment == self._include_pages)
            ):
                continue

            record = self._get(name, persist=False)
            if self._matches(record):
                yield record

    def filter(self, expr):
        """Filters records by an expression."""
        rv = self._clone(mark_dirty=True)
        rv._filters = list(self._filters or ())
        if callable(expr):
            expr = _CallbackExpr(expr)
        rv._filters.append(expr)
        return rv

    def get_order_by(self):
        """Returns the order that should be used."""
        if self._order_by is not None:
            return self._order_by
        base_record = self.pad.get(self.path)
        if base_record is not None:
            if self._include_attachments and not self._include_pages:
                return base_record.datamodel.attachment_config.order_by
            if self._include_pages and not self._include_attachments:
                return base_record.datamodel.child_config.order_by
            # Otherwise the query includes either both or neither
            # attachments and/nor children.  I have no idea which
            # value of order_by to use.  We could punt and return
            # child_config.order_by, but for now, just return None.
            return None
        return None

    def include_hidden(self, value):
        """Controls if hidden records should be included which will not
        happen by default for queries to children.
        """
        rv = self._clone(mark_dirty=True)
        rv._include_hidden = value
        return rv

    def include_undiscoverable(self, value):
        """Controls if undiscoverable records should be included as well."""
        rv = self._clone(mark_dirty=True)
        rv._include_undiscoverable = value

        # If we flip from not including undiscoverables to discoverables
        # but we did not yet decide on the value of _include_hidden it
        # becomes False to not include it.
        if rv._include_hidden is None and value:
            rv._include_hidden = False

        return rv

    def request_page(self, page_num):
        """Requests a specific page number instead of the first."""
        rv = self._clone(mark_dirty=True)
        rv._page_num = page_num
        return rv

    def first(self):
        """Return the first matching record."""
        return next(iter(self), None)

    def all(self):
        """Loads all matching records as list."""
        return list(self)

    def order_by(self, *fields):
        """Sets the ordering of the query."""
        rv = self._clone()
        rv._order_by = fields or None
        return rv

    def offset(self, offset):
        """Sets the ordering of the query."""
        rv = self._clone(mark_dirty=True)
        rv._offset = offset
        return rv

    def limit(self, limit):
        """Sets the ordering of the query."""
        rv = self._clone(mark_dirty=True)
        rv._limit = limit
        return rv

    def count(self):
        """Counts all matched objects."""
        rv = 0
        for _ in self:
            rv += 1
        return rv

    def distinct(self, fieldname):
        """Set of unique values for the given field."""
        rv = set()

        for item in self:
            if fieldname in item._data:
                value = item._data[fieldname]
                if isinstance(value, (list, tuple)):
                    rv |= set(value)
                elif not isinstance(value, Undefined):
                    rv.add(value)

        return rv

    def get(self, id, page_num=Ellipsis):
        """Gets something by the local path."""
        # If we're not pristine, we need to query here
        if not self._pristine:
            q = self.filter(F._id == id)
            if page_num is not Ellipsis:
                q = q.request_page(page_num)
            return q.first()
        # otherwise we can load it directly.
        return self._get(id, page_num=page_num)

    def __bool__(self):
        return self.first() is not None

    __nonzero__ = __bool__

    def __iter__(self):
        """Iterates over all records matched."""
        iterable = self._iterate()

        order_by = self.get_order_by()
        if order_by:
            iterable = sorted(iterable, key=lambda x: x.get_sort_key(order_by))

        if self._offset is not None or self._limit is not None:
            iterable = islice(
                iterable,
                self._offset or 0,
                (self._offset or 0) + self._limit if self._limit else None,
            )

        for item in iterable:
            yield item

    def __repr__(self):
        return "<%s %r%s>" % (
            self.__class__.__name__,
            self.path,
            self.alt and " alt=%r" % self.alt or "",
        )


class EmptyQuery(Query):
    def _get(self, id, persist=True, page_num=Ellipsis):
        pass

    def _iterate(self):
        """Low level record iteration."""
        return iter(())


class AttachmentsQuery(Query):
    """Specialized query class that only finds attachments."""

    def __init__(self, path, pad, alt=PRIMARY_ALT):
        Query.__init__(self, path, pad, alt=alt)
        self._include_pages = False
        self._include_attachments = True

    @property
    def images(self):
        """Filters to images."""
        return self.filter(F._attachment_type == "image")

    @property
    def videos(self):
        """Filters to videos."""
        return self.filter(F._attachment_type == "video")

    @property
    def audio(self):
        """Filters to audio."""
        return self.filter(F._attachment_type == "audio")

    @property
    def documents(self):
        """Filters to documents."""
        return self.filter(F._attachment_type == "document")

    @property
    def text(self):
        """Filters to plain text data."""
        return self.filter(F._attachment_type == "text")


def _iter_filename_choices(fn_base, alts, config, fallback=True):
    """Returns an iterator over all possible filename choices to .lr files
    below a base filename that matches any of the given alts.
    """
    # the order here is important as attachments can exist without a .lr
    # file and as such need to come second or the loading of raw data will
    # implicitly say the record exists.
    for alt in alts:
        if alt != PRIMARY_ALT and config.is_valid_alternative(alt):
            yield os.path.join(fn_base, "contents+%s.lr" % alt), alt, False

    if fallback or PRIMARY_ALT in alts:
        yield os.path.join(fn_base, "contents.lr"), PRIMARY_ALT, False

    for alt in alts:
        if alt != PRIMARY_ALT and config.is_valid_alternative(alt):
            yield fn_base + "+%s.lr" % alt, alt, True

    if fallback or PRIMARY_ALT in alts:
        yield fn_base + ".lr", PRIMARY_ALT, True


def _iter_content_files(dir_path, alts):
    """Returns an iterator over all existing content files below the given
    directory.  This yields specific files for alts before it falls back
    to the primary alt.
    """
    for alt in alts:
        if alt == PRIMARY_ALT:
            continue
        if os.path.isfile(os.path.join(dir_path, "contents+%s.lr" % alt)):
            yield alt
    if os.path.isfile(os.path.join(dir_path, "contents.lr")):
        yield PRIMARY_ALT


def _iter_datamodel_choices(datamodel_name, path, is_attachment=False):
    yield datamodel_name
    if not is_attachment:
        yield posixpath.basename(path).split(".")[0].replace("-", "_").lower()
        yield "page"
    yield "none"


def get_default_slug(record):
    """Compute the default slug for a page.

    This computes the default value of ``_slug`` for a page.  The slug
    is computed by expanding the parent’s ``slug_format`` value.

    """
    parent = getattr(record, "parent", None)
    if parent is None:
        return ""
    return parent.datamodel.get_default_child_slug(record.pad, record)


default_slug_descriptor = property(get_default_slug)


class Database:
    def __init__(self, env, config=None):
        self.env = env
        if config is None:
            config = env.load_config()
        self.config = config
        self.datamodels = load_datamodels(env)
        self.flowblocks = load_flowblocks(env)

    def to_fs_path(self, path):
        """Convenience function to convert a path into an file system path."""
        return os.path.join(self.env.root_path, "content", untrusted_to_os_path(path))

    def load_raw_data(self, path, alt=PRIMARY_ALT, cls=None, fallback=True):
        """Internal helper that loads the raw record data.  This performs
        very little data processing on the data.
        """
        path = cleanup_path(path)
        if cls is None:
            cls = dict

        fn_base = self.to_fs_path(path)

        rv = cls()
        rv_type = None

        choiceiter = _iter_filename_choices(
            fn_base, [alt], self.config, fallback=fallback
        )
        for fs_path, source_alt, is_attachment in choiceiter:
            # If we already determined what our return value is but the
            # type mismatches what we try now, we have to abort.  Eg:
            # a page can not become an attachment or the other way round.
            if rv_type is not None and rv_type != is_attachment:
                break

            try:
                with open(fs_path, "rb") as f:
                    if rv_type is None:
                        rv_type = is_attachment
                    for key, lines in metaformat.tokenize(f, encoding="utf-8"):
                        if key not in rv:
                            rv[key] = "".join(lines)
            except OSError as e:
                if e.errno not in (errno.ENOTDIR, errno.ENOENT, errno.EINVAL):
                    raise
                if not is_attachment or not os.path.isfile(fs_path[:-3]):
                    continue
                # Special case: we are loading an attachment but the meta
                # data file does not exist.  In that case we still want to
                # record that we're loading an attachment.
                if is_attachment:
                    rv_type = True

            if "_source_alt" not in rv:
                rv["_source_alt"] = source_alt

        if rv_type is None:
            return None

        rv["_path"] = path
        rv["_id"] = posixpath.basename(path)
        rv["_gid"] = hashlib.md5(path.encode("utf-8")).hexdigest()
        rv["_alt"] = alt
        if rv_type:
            rv["_attachment_for"] = posixpath.dirname(path)

        return rv

    def iter_items(self, path, alt=PRIMARY_ALT):
        """Iterates over all items below a path and yields them as
        tuples in the form ``(id, alt, is_attachment)``.
        """
        fn_base = self.to_fs_path(path)

        if alt is None:
            alts = self.config.list_alternatives()
            single_alt = False
        else:
            alts = [alt]
            single_alt = True

        choiceiter = _iter_filename_choices(fn_base, alts, self.config)

        for fs_path, _actual_alt, is_attachment in choiceiter:
            if not os.path.isfile(fs_path):
                continue

            # This path is actually for an attachment, which means that we
            # cannot have any items below it and will just abort with an
            # empty iterator.
            if is_attachment:
                break

            try:
                dir_path = os.path.dirname(fs_path)
                for filename in os.listdir(dir_path):
                    if not isinstance(filename, str):
                        try:
                            filename = filename.decode(fs_enc)
                        except UnicodeError:
                            continue

                    if filename.endswith(
                        ".lr"
                    ) or self.env.is_uninteresting_source_name(filename):
                        continue

                    # We found an attachment.  Attachments always live
                    # below the primary alt, so we report it as such.
                    if os.path.isfile(os.path.join(dir_path, filename)):
                        yield filename, PRIMARY_ALT, True

                    # We found a directory, let's make sure it contains a
                    # contents.lr file (or a contents+alt.lr file).
                    else:
                        for content_alt in _iter_content_files(
                            os.path.join(dir_path, filename), alts
                        ):
                            yield filename, content_alt, False
                            # If we want a single alt, we break here so
                            # that we only produce a single result.
                            # Otherwise this would also return the primary
                            # fallback here.
                            if single_alt:
                                break
            except IOError as e:
                if e.errno != errno.ENOENT:
                    raise
                continue

            # If we reach this point, we found our parent, so we can stop
            # searching for more at this point.
            break

    def get_datamodel_for_raw_data(self, raw_data, pad=None):
        """Returns the datamodel that should be used for a specific raw
        data.  This might require the discovery of a parent object through
        the pad.
        """
        path = raw_data["_path"]
        is_attachment = bool(raw_data.get("_attachment_for"))
        datamodel = (raw_data.get("_model") or "").strip() or None
        return self.get_implied_datamodel(path, is_attachment, pad, datamodel=datamodel)

    def iter_dependent_models(self, datamodel):
        seen = set()

        def deep_find(datamodel):
            seen.add(datamodel)

            if datamodel.parent is not None and datamodel.parent not in seen:
                deep_find(datamodel.parent)

            for related_dm_name in (
                datamodel.child_config.model,
                datamodel.attachment_config.model,
            ):
                dm = self.datamodels.get(related_dm_name)
                if dm is not None and dm not in seen:
                    deep_find(dm)

        deep_find(datamodel)
        seen.discard(datamodel)
        return iter(seen)

    def get_implied_datamodel(
        self, path, is_attachment=False, pad=None, datamodel=None
    ):
        """Looks up a datamodel based on the information about the parent
        of a model.
        """
        dm_name = datamodel

        # Only look for a datamodel if there was not defined.
        if dm_name is None:
            parent = posixpath.dirname(path)
            dm_name = None

            # If we hit the root, and there is no model defined we need
            # to make sure we do not recurse onto ourselves.
            if parent != path:
                if pad is None:
                    pad = self.new_pad()
                parent_obj = pad.get(parent)
                if parent_obj is not None:
                    if is_attachment:
                        dm_name = parent_obj.datamodel.attachment_config.model
                    else:
                        dm_name = parent_obj.datamodel.child_config.model

        for dm_name in _iter_datamodel_choices(dm_name, path, is_attachment):
            # If that datamodel exists, let's roll with it.
            datamodel = self.datamodels.get(dm_name)
            if datamodel is not None:
                return datamodel

        raise AssertionError(
            "Did not find an appropriate datamodel.  " "That should never happen."
        )

    def get_attachment_type(self, path):
        """Gets the attachment type for a path."""
        return self.config["ATTACHMENT_TYPES"].get(posixpath.splitext(path)[1].lower())

    def track_record_dependency(self, record):
        ctx = get_ctx()
        if ctx is not None:
            for filename in record.iter_source_filenames():
                ctx.record_dependency(filename)
            for virtual_source in record.iter_virtual_sources():
                ctx.record_virtual_dependency(virtual_source)
            if getattr(record, "datamodel", None) and record.datamodel.filename:
                ctx.record_dependency(record.datamodel.filename)
                for dep_model in self.iter_dependent_models(record.datamodel):
                    if dep_model.filename:
                        ctx.record_dependency(dep_model.filename)
            # XXX: In the case that our datamodel is implied, then the
            # datamodel depends on the datamodel(s) of our parent(s).
            # We do not currently record that.
        return record

    def process_data(self, data, datamodel, pad):
        # Automatically fill in slugs
        if not data["_slug"]:
            data["_slug"] = default_slug_descriptor
        else:
            data["_slug"] = data["_slug"].strip("/")

        # For attachments figure out the default attachment type if it's
        # not yet provided.
        if is_undefined(data["_attachment_type"]) and data["_attachment_for"]:
            data["_attachment_type"] = self.get_attachment_type(data["_path"])

        # Automatically fill in templates
        if is_undefined(data["_template"]):
            data["_template"] = datamodel.get_default_template_name()

    @staticmethod
    def get_record_class(datamodel, raw_data):
        """Returns the appropriate record class for a datamodel and raw data."""
        is_attachment = bool(raw_data.get("_attachment_for"))
        if not is_attachment:
            return Page
        attachment_type = raw_data["_attachment_type"]
        return attachment_classes.get(attachment_type, Attachment)

    def new_pad(self):
        """Creates a new pad object for this database."""
        return Pad(self)


def _split_alt_from_url(config, clean_path):
    primary = config.primary_alternative

    # The alternative system is not configured, just return
    if primary is None:
        return None, clean_path

    # First try to find alternatives that are identified by a prefix.
    for prefix, alt in config.get_alternative_url_prefixes():
        if clean_path.startswith(prefix):
            return alt, clean_path[len(prefix) :].strip("/")
        # Special case which is the URL root.
        if prefix.strip("/") == clean_path:
            return alt, ""

    # Now find alternatives that are identified by a suffix.
    for suffix, alt in config.get_alternative_url_suffixes():
        if clean_path.endswith(suffix):
            return alt, clean_path[: -len(suffix)].strip("/")

    # If we have a primary alternative without a prefix and suffix, we can
    # return that one.
    if config.primary_alternative_is_rooted:
        return None, clean_path

    return None, None


class Pad:
    def __init__(self, db):
        self.db = db
        self.cache = RecordCache(db.config["EPHEMERAL_RECORD_CACHE_SIZE"])
        self.databags = Databags(db.env)

    @property
    def config(self):
        """The config for this pad."""
        return self.db.config

    @property
    def env(self):
        """The env for this pad."""
        return self.db.env

    def make_absolute_url(self, url):
        """Given a URL this makes it absolute if this is possible."""
        base_url = self.db.config["PROJECT"].get("url")
        if base_url is None:
            raise RuntimeError(
                "To use absolute URLs you need to configure "
                "the URL in the project config."
            )
        return url_join(base_url.rstrip("/") + "/", url.lstrip("/"))

    def make_url(self, url, base_url=None, absolute=None, external=None):
        """Helper method that creates a finalized URL based on the parameters
        provided and the config.
        """
        url_style = self.db.config.url_style
        if absolute is None:
            absolute = url_style == "absolute"
        if external is None:
            external = url_style == "external"
        if external:
            external_base_url = self.db.config.base_url
            if external_base_url is None:
                raise RuntimeError(
                    "To use absolute URLs you need to "
                    "configure the URL in the project config."
                )
            return url_join(external_base_url, url.lstrip("/"))
        if absolute:
            return url_join(self.db.config.base_path, url.lstrip("/"))
        if base_url is None:
            raise RuntimeError(
                "Cannot calculate a relative URL if no base " "URL has been provided."
            )
        return make_relative_url(base_url, url)

    def resolve_url_path(
        self, url_path, include_invisible=False, include_assets=True, alt_fallback=True
    ):
        """Given a URL path this will find the correct record which also
        might be an attachment.  If a record cannot be found or is unexposed
        the return value will be `None`.
        """
        pieces = clean_path = cleanup_path(url_path).strip("/")

        # Split off the alt and if no alt was found, point it to the
        # primary alternative.  If the clean path comes back as `None`
        # then the config does not include a rooted alternative and we
        # have to skip handling of regular records.
        alt, clean_path = _split_alt_from_url(self.db.config, clean_path)
        if clean_path is not None:
            if not alt:
                if alt_fallback:
                    alt = self.db.config.primary_alternative or PRIMARY_ALT
                else:
                    alt = PRIMARY_ALT
            node = self.get_root(alt=alt)
            if node is None:
                raise RuntimeError("Tree root could not be found.")

            pieces = clean_path.split("/")
            if pieces == [""]:
                pieces = []

            rv = node.resolve_url_path(pieces)
            if rv is not None and (include_invisible or rv.is_visible):
                return rv

        if include_assets:
            for asset_root in [self.asset_root] + self.theme_asset_roots:
                rv = asset_root.resolve_url_path(pieces)
                if rv is not None:
                    break
            return rv
        return None

    def get_root(self, alt=None):
        """The root page of the database."""
        if alt is None:
            alt = self.config.primary_alternative or PRIMARY_ALT
        return self.get("/", alt=alt, persist=True)

    root = property(get_root)

    @property
    def asset_root(self):
        """The root of the asset tree."""
        return Directory(
            self, name="", path=os.path.join(self.db.env.root_path, "assets")
        )

    @property
    def theme_asset_roots(self):
        """The root of the asset tree of each theme."""
        asset_roots = []
        for theme_path in self.db.env.theme_paths:
            asset_roots.append(
                Directory(self, name="", path=os.path.join(theme_path, "assets"))
            )
        return asset_roots

    def get_all_roots(self):
        """Returns all the roots for building."""
        rv = []
        for alt in self.db.config.list_alternatives():
            rv.append(self.get_root(alt=alt))

        # If we don't have any alternatives, then we go with the implied
        # root.
        if not rv and self.root:
            rv = [self.root]

        rv.append(self.asset_root)
        rv.extend(self.theme_asset_roots)
        return rv

    def get_virtual(self, record, virtual_path):
        """Resolves a virtual path below a record."""
        pieces = virtual_path.strip("/").split("/")
        if not pieces or pieces == [""]:
            return record

        if pieces[0].isdigit():
            if len(pieces) == 1:
                return self.get(
                    record["_path"], alt=record.alt, page_num=int(pieces[0])
                )
            return None

        resolver = self.env.virtual_sources.get(pieces[0])
        if resolver is None:
            return None

        return resolver(record, pieces[1:])

    def get(self, path, alt=None, page_num=None, persist=True, allow_virtual=True):
        """Loads a record by path."""
        if alt is None:
            alt = self.config.primary_alternative or PRIMARY_ALT
        virt_markers = path.count("@")

        # If the virtual marker is included, we also want to look up the
        # virtual path below an item.  Special case: if virtual paths are
        # not allowed but one was passed, we just return `None`.
        if virt_markers == 1:
            if page_num is not None:
                raise RuntimeError(
                    "Cannot use both virtual paths and "
                    "explicit page number lookups.  You "
                    "need to one or the other."
                )
            if not allow_virtual:
                return None
            path, virtual_path = path.split("@", 1)
            rv = self.get(path, alt=alt, page_num=page_num, persist=persist)
            if rv is None:
                return None
            return self.get_virtual(rv, virtual_path)

        # Sanity check: there must only be one or things will get weird.
        if virt_markers > 1:
            return None

        path = cleanup_path(path)
        virtual_path = None
        if page_num is not None:
            virtual_path = str(page_num)

        rv = self.cache.get(path, alt, virtual_path)
        if rv is not Ellipsis:
            if rv is not None:
                self.db.track_record_dependency(rv)
            return rv

        raw_data = self.db.load_raw_data(path, alt=alt)
        if raw_data is None:
            self.cache.remember_as_missing(path, alt, virtual_path)
            return None

        rv = self.instance_from_data(raw_data, page_num=page_num)

        if persist:
            self.cache.persist(rv)
        else:
            self.cache.remember(rv)

        return self.db.track_record_dependency(rv)

    def alt_exists(self, path, alt=PRIMARY_ALT, fallback=False):
        """Checks if an alt exists."""
        path = cleanup_path(path)
        if "@" in path:
            return False

        # If we find the path in the cache, check if it was loaded from
        # the right source alt.
        rv = self.get(path, alt)
        if rv is not None:
            if rv["_source_alt"] == alt:
                return True
            if fallback or (
                rv["_source_alt"] == PRIMARY_ALT
                and alt == self.config.primary_alternative
            ):
                return True
            return False

        return False

    def get_asset(self, path):
        """Loads an asset by path."""
        clean_path = cleanup_path(path).strip("/")
        nodes = [self.asset_root] + self.theme_asset_roots
        for node in nodes:
            for piece in clean_path.split("/"):
                node = node.get_child(piece)
                if node is None:
                    break
            if node is not None:
                return node
        return None

    def instance_from_data(self, raw_data, datamodel=None, page_num=None):
        """This creates an instance from the given raw data."""
        if datamodel is None:
            datamodel = self.db.get_datamodel_for_raw_data(raw_data, self)
        data = datamodel.process_raw_data(raw_data, self)
        self.db.process_data(data, datamodel, self)
        cls = self.db.get_record_class(datamodel, data)
        return cls(self, data, page_num=page_num)

    def query(self, path=None, alt=PRIMARY_ALT):
        """Queries the database either at root level or below a certain
        path.  This is the recommended way to interact with toplevel data.
        The alternative is to work with the :attr:`root` document.
        """
        # Don't accidentally pass `None` down to the query as this might
        # do some unexpected things.
        if alt is None:
            alt = PRIMARY_ALT
        return Query(
            path="/" + (path or "").strip("/"), pad=self, alt=alt
        ).include_hidden(True)


class TreeItem:
    """Represents a single tree item and all the alts within it."""

    def __init__(self, tree, path, alts, primary_record):
        self.tree = tree
        self.path = path
        self.alts = alts
        self._primary_record = primary_record

    @property
    def id(self):
        """The local ID of the item."""
        return posixpath.basename(self.path)

    @property
    def exists(self):
        """True iff metadata exists for the item.

        If metadata exists for any alt, including the fallback (PRIMARY_ALT).

        Note that for attachments without metadata, this is currently False.
        But note that if is_attachment is True, the attachment file does exist.
        """
        # FIXME: this should probably be changed to return True for attachments,
        # even those without metadata.
        return self._primary_record is not None

    @property
    def is_attachment(self):
        """True iff item is an attachment."""
        if self._primary_record is None:
            return False
        return self._primary_record.is_attachment

    @property
    def is_visible(self):
        """True iff item is not hidden."""
        # XXX: This is send out from /api/recordinfo but appears to be
        # unused by the React app
        if self._primary_record is None:
            return True
        return self._primary_record.is_visible

    @property
    def can_be_deleted(self):
        """True iff item can be deleted."""
        if self.path == "/" or not self.exists:
            return False
        return self.is_attachment or not self._datamodel.protected

    @property
    def _datamodel(self):
        if self._primary_record is None:
            return None
        return self._primary_record.datamodel

    def get_record_label_i18n(self, alt=PRIMARY_ALT):
        """Get record label translations for specific alt."""
        record = self.alts[alt].record
        if record is None:
            # generate a reasonable fallback
            # ("en" is the magical fallback lang)
            label = self.id.replace("-", " ").replace("_", " ").title()
            return {"en": label or "(Index)"}
        return record.get_record_label_i18n()

    @property
    def can_have_children(self):
        """True iff the item can contain subpages."""
        if self._primary_record is None or self.is_attachment:
            return False
        return self._datamodel.has_own_children

    @property
    def implied_child_datamodel(self):
        """The name of the default datamodel for children of this page, if any."""
        datamodel = self._datamodel
        return datamodel.child_config.model if datamodel else None

    @property
    def can_have_attachments(self):
        """True iff the item can contain attachments."""
        if self._primary_record is None or self.is_attachment:
            return False
        return self._datamodel.has_own_attachments

    @property
    def attachment_type(self):
        """The type of an attachment.

        E.g. "image", "video", or None if type is unknown.
        """
        if self._primary_record is None or not self.is_attachment:
            return None
        return self._primary_record["_attachment_type"] or None

    def get_parent(self):
        """Returns the parent item."""
        if self.path == "/":
            return None
        return self.tree.get(posixpath.dirname(self.path))

    def get(self, path):
        """Returns a child within this item."""
        # XXX: Unused?
        return self.tree.get(posixpath.join(self.path, path))

    def _get_child_ids(self, include_attachments=True, include_pages=True):
        """Returns a sorted list of just the IDs of existing children."""
        db = self.tree.pad.db
        keep_attachments = include_attachments and self.can_have_attachments
        keep_pages = include_pages and self.can_have_children
        names = set(
            name
            for name, _, is_attachment in db.iter_items(self.path, alt=None)
            if (keep_attachments if is_attachment else keep_pages)
        )
        return sorted(names, key=lambda name: name.lower())

    def iter_children(
        self, include_attachments=True, include_pages=True, order_by=None
    ):
        """Iterates over all children"""
        children = (
            self.tree.get(posixpath.join(self.path, name), persist=False)
            for name in self._get_child_ids(include_attachments, include_pages)
        )
        if order_by is not None:
            children = sorted(children, key=methodcaller("get_sort_key", order_by))
        return children

    def get_children(
        self,
        offset=0,
        limit=None,
        include_attachments=True,
        include_pages=True,
        order_by=None,
    ):
        """Returns a slice of children."""
        # XXX: this method appears unused?
        end = None
        if limit is not None:
            end = offset + limit
        children = self.iter_children(include_attachments, include_pages, order_by)
        return list(islice(children, offset, end))

    def iter_attachments(self, order_by=None):
        """Return an iterable of this records attachments.

        By default, the attachments are sorted as specified by
        ``[attachments]order_by`` in this records datamodel.
        """
        if order_by is None:
            dm = self._datamodel
            if dm is not None:
                order_by = dm.attachment_config.order_by
        return self.iter_children(include_pages=False, order_by=order_by)

    def iter_subpages(self, order_by=None):
        """Return an iterable of this records sub-pages.

        By default, the records are sorted as specified by
        ``[children]order_by`` in this records datamodel.

        NB: This method should probably be called ``iter_children``,
        but that name was already taken.
        """
        if order_by is None:
            dm = self._datamodel
            if dm is not None:
                order_by = dm.child_config.order_by
        return self.iter_children(include_attachments=False, order_by=order_by)

    def get_sort_key(self, order_by):
        if self._primary_record is None:

            def sort_key(fieldspec):
                if fieldspec.startswith("-"):
                    field, reverse = fieldspec[1:], True
                else:
                    field, reverse = fieldspec.lstrip("+"), False
                value = self.id if field == "_id" else None
                return _CmpHelper(value, reverse)

            return [sort_key(fieldspec) for fieldspec in order_by]
        return self._primary_record.get_sort_key(order_by)

    def __repr__(self):
        return "<TreeItem %r%s>" % (
            self.path,
            self.is_attachment and " attachment" or "",
        )


class Alt:
    def __init__(self, id, record, is_primary_overlay, name_i18n):
        self.id = id
        self.record = record
        self.is_primary_overlay = is_primary_overlay
        self.name_i18n = name_i18n
        self.exists = record is not None and os.path.isfile(record.source_filename)

    def __repr__(self):
        return "<Alt %r%s>" % (self.id, self.exists and "*" or "")


class Tree:
    """Special object that can be used to get a broader insight into the
    database in a way that is not bound to the alt system directly.

    This wraps a pad and provides additional ways to interact with the data
    of the database in a way that is similar to how the data is actually laid
    out on the file system and not as the data is represented.  Primarily the
    difference is how alts are handled.  Where the pad resolves the alts
    automatically to make the handling automatic, the tree will give access
    to the underlying alts automatically.
    """

    def __init__(self, pad):
        # Note that in theory different alts can disagree on what
        # datamodel they use but this is something that is really not
        # supported.  This cannot happen if you edit based on the admin
        # panel and if you edit it manually and screw up that part, we
        # cannot really do anything about it.
        #
        # We will favor the datamodel from the fallback record
        # (alt=PRIMARY_ALT) if it exists. If it does not, we will
        # use the data for the primary alternative. If data does not exist
        # for either of those, we will pick from among the other
        # existing alts quasi-randomly.
        #
        # Here we construct a list of configured alts in preference order
        config = pad.db.config
        alt_info = OrderedDict()
        if config.primary_alternative:
            # Alternatives are configured
            alts = [PRIMARY_ALT]
            alts.append(config.primary_alternative)
            alts.extend(
                alt
                for alt in config.list_alternatives()
                if alt != config.primary_alternative
            )
            for alt in alts:
                alt_info[alt] = {
                    "is_primary_overlay": alt == config.primary_alternative,
                    "name_i18n": config.get_alternative(alt)["name"],
                }
        else:
            alt_info[PRIMARY_ALT] = {
                "is_primary_overlay": True,
                "name_i18n": {"en": "Primary"},
            }

        self.pad = pad
        self._alt_info = alt_info

    def get(self, path=None, persist=True):
        """Returns a path item at the given node."""
        path = "/" + (path or "").strip("/")
        alts = {}
        primary_record = None
        for alt, alt_info in self._alt_info.items():
            record = self.pad.get(path, alt=alt, persist=persist, allow_virtual=False)
            if primary_record is None:
                primary_record = record
            alts[alt] = Alt(alt, record, **alt_info)
        return TreeItem(self, path, alts, primary_record)

    def iter_children(
        self, path=None, include_attachments=True, include_pages=True, order_by=None
    ):
        """Iterates over all children below a path"""
        # XXX: this method is unused?
        path = "/" + (path or "").strip("/")
        return self.get(path, persist=False).iter_children(
            include_attachments, include_pages, order_by
        )

    def get_children(
        self,
        path=None,
        offset=0,
        limit=None,
        include_attachments=True,
        include_pages=True,
        order_by=None,
    ):
        """Returns a slice of children."""
        # XXX: this method is unused?
        path = "/" + (path or "").strip("/")
        return self.get(path, persist=False).get_children(
            offset, limit, include_attachments, include_pages, order_by
        )

    def edit(self, path, is_attachment=None, alt=PRIMARY_ALT, datamodel=None):
        """Edits a record by path."""
        return make_editor_session(
            self.pad,
            cleanup_path(path),
            alt=alt,
            is_attachment=is_attachment,
            datamodel=datamodel,
        )


class RecordCache:
    """The record cache holds records eitehr in an persistent or ephemeral
    section which helps the pad not load records it already saw.
    """

    def __init__(self, ephemeral_cache_size=1000):
        self.persistent = {}
        self.ephemeral = LRUCache(ephemeral_cache_size)

    @staticmethod
    def _get_cache_key(record_or_path, alt=PRIMARY_ALT, virtual_path=None):
        if isinstance(record_or_path, str):
            path = record_or_path.strip("/")
        else:
            path, virtual_path = split_virtual_path(record_or_path.path)
            path = path.strip("/")
            virtual_path = virtual_path or None
            alt = record_or_path.alt
        return (path, alt, virtual_path)

    def flush(self):
        """Flushes the cache"""
        self.persistent.clear()
        self.ephemeral.clear()

    def is_persistent(self, record):
        """Indicates if a record is in the persistent record cache."""
        cache_key = self._get_cache_key(record)
        return cache_key in self.persistent

    def remember(self, record):
        """Remembers the record in the record cache."""
        cache_key = self._get_cache_key(record)
        if cache_key not in self.persistent and cache_key not in self.ephemeral:
            self.ephemeral[cache_key] = record

    def persist(self, record):
        """Persists a record.  This will put it into the persistent cache."""
        cache_key = self._get_cache_key(record)
        self.persistent[cache_key] = record
        try:
            del self.ephemeral[cache_key]
        except KeyError:
            pass

    def persist_if_cached(self, record):
        """If the record is already ephemerally cached, this promotes it to
        the persistent cache section.
        """
        cache_key = self._get_cache_key(record)
        if cache_key in self.ephemeral:
            self.persist(record)

    def get(self, path, alt=PRIMARY_ALT, virtual_path=None):
        """Looks up a record from the cache."""
        cache_key = self._get_cache_key(path, alt, virtual_path)
        rv = self.persistent.get(cache_key, Ellipsis)
        if rv is not Ellipsis:
            return rv
        rv = self.ephemeral.get(cache_key, Ellipsis)
        if rv is not Ellipsis:
            return rv
        return Ellipsis

    def remember_as_missing(self, path, alt=PRIMARY_ALT, virtual_path=None):
        cache_key = self._get_cache_key(path, alt, virtual_path)
        self.persistent.pop(cache_key, None)
        self.ephemeral[cache_key] = None



# coding: utf-8
import pytest

from lektor.utils import build_url
from lektor.utils import is_path_child_of
from lektor.utils import join_path
from lektor.utils import magic_split_ext
from lektor.utils import make_relative_url
from lektor.utils import parse_path
from lektor.utils import slugify


def test_join_path():

    assert join_path("a", "b") == "a/b"
    assert join_path("/a", "b") == "/a/b"
    assert join_path("a@b", "c") == "a@b/c"
    assert join_path("a@b", "") == "a@b"
    assert join_path("a@b", "@c") == "a@c"
    assert join_path("a@b/c", "a@b") == "a/a@b"

    assert join_path("blog@archive", "2015") == "blog@archive/2015"
    assert join_path("blog@archive/2015", "..") == "blog@archive"
    assert join_path("blog@archive/2015", "@archive") == "blog@archive"
    assert join_path("blog@archive", "..") == "blog"
    assert join_path("blog@archive", ".") == "blog@archive"
    assert join_path("blog@archive", "") == "blog@archive"

    # special behavior: parent of pagination paths is always the actual
    # page parent.
    assert join_path("/blog@1", "..") == "/"
    assert join_path("/blog@2", "..") == "/"

    # But joins on the same level keep the path
    assert join_path("/blog@1", ".") == "/blog@1"
    assert join_path("/blog@2", ".") == "/blog@2"
    assert join_path("/blog@1", "") == "/blog@1"
    assert join_path("/blog@2", "") == "/blog@2"


def test_is_path_child_of():

    assert not is_path_child_of("a/b", "a/b")
    assert is_path_child_of("a/b", "a/b", strict=False)
    assert is_path_child_of("a/b/c", "a")
    assert not is_path_child_of("a/b/c", "b")
    assert is_path_child_of("a/b@foo/bar", "a/b@foo")
    assert is_path_child_of("a/b@foo", "a/b@foo", strict=False)
    assert not is_path_child_of("a/b@foo/bar", "a/c@foo")
    assert not is_path_child_of("a/b@foo/bar", "a/c")
    assert is_path_child_of("a/b@foo", "a/b")
    assert is_path_child_of("a/b@foo/bar", "a/b@foo")
    assert not is_path_child_of("a/b@foo/bar", "a/b@bar")


def test_magic_split_ext():

    assert magic_split_ext("wow") == ("wow", "")
    assert magic_split_ext("aaa.jpg") == ("aaa", "jpg")
    assert magic_split_ext("aaa. jpg") == ("aaa. jpg", "")
    assert magic_split_ext("aaa.j pg") == ("aaa.j pg", "")
    assert magic_split_ext("aaa.j pg", ext_check=False) == ("aaa", "j pg")


def test_slugify():

    assert slugify("w o w") == "w-o-w"
    assert slugify("Șö prĕtty") == "so-pretty"
    assert slugify("im age.jpg") == "im-age.jpg"
    assert slugify("slashed/slug") == "slashed/slug"


def test_url_builder():

    assert build_url([]) == "/"
    assert build_url(["a", "b/c"]) == "/a/b/c/"
    assert build_url(["a", "b/c"], trailing_slash=False) == "/a/b/c"
    assert build_url(["a", "b/c.html"]) == "/a/b/c.html"
    assert build_url(["a", "b/c.html"], trailing_slash=True) == "/a/b/c.html/"
    assert build_url(["a", None, "b", "", "c"]) == "/a/b/c/"


def test_parse_path():

    assert parse_path("") == []
    assert parse_path("/") == []
    assert parse_path("/foo") == ["foo"]
    assert parse_path("/foo/") == ["foo"]
    assert parse_path("/foo/bar") == ["foo", "bar"]
    assert parse_path("/foo/bar/") == ["foo", "bar"]
    assert parse_path("/foo/bar/../stuff") == ["foo", "bar", "stuff"]


@pytest.mark.parametrize(
    "source, target, expected",
    [
        ("/", "./a/", "a/"),
        ("/", "./a", "a"),
        ("/fr/blog/2015/11/a/", "/fr/blog/2015/11/a/a.jpg", "a.jpg"),
        ("/fr/blog/2015/11/a/", "/fr/blog/", "../../../"),
        ("/fr/blog/2015/11/a.php", "/fr/blog/", "../../"),
        ("/fr/blog/2015/11/a/", "/fr/blog/2016/", "../../../2016/"),
        ("/fr/blog/2015/11/a/", "/fr/blog/2016/c.jpg", "../../../2016/c.jpg"),
        ("/fr/blog/2016/", "/fr/blog/2015/a/", "../2015/a/"),
        ("/fr/blog/2016/", "/fr/blog/2015/a/d.jpg", "../2015/a/d.jpg"),
        ("/fr/blog/2015/11/a/", "/images/b.svg", "../../../../../images/b.svg"),
        ("/fr/blog/", "2015/11/", "2015/11/"),
        ("/fr/blog/x", "2015/11/", "2015/11/"),
        ("", "./a/", "a/"),
        ("", "./a", "a"),
        ("fr/blog/2015/11/a/", "fr/blog/2015/11/a/a.jpg", "a.jpg"),
        ("fr/blog/2015/11/a/", "fr/blog/", "../../../"),
        ("fr/blog/2015/11/a.php", "fr/blog/", "../../"),
        ("fr/blog/2015/11/a/", "fr/blog/2016/", "../../../2016/"),
        ("fr/blog/2015/11/a/", "fr/blog/2016/c.jpg", "../../../2016/c.jpg"),
        ("fr/blog/2016/", "fr/blog/2015/a/", "../2015/a/"),
        ("fr/blog/2016/", "fr/blog/2015/a/d.jpg", "../2015/a/d.jpg"),
        ("fr/blog/2015/11/a/", "images/b.svg", "../../../../../images/b.svg"),
        ("fr/blog/", "2015/11/", "../../2015/11/"),
        ("fr/blog/x", "2015/11/", "../../2015/11/"),
    ],
)
def test_make_relative_url(source, target, expected):
    assert make_relative_url(source, target) == expected


def test_make_relative_url_relative_source_absolute_target():
    with pytest.raises(ValueError):
        make_relative_url("rel/a/tive/", "/abs/o/lute")



from functools import update_wrapper
from itertools import chain

from flask import json
from flask import Response


def fs_path_to_url_path(path):
    segments = path.strip("/").split("/")
    if segments == [""]:
        segments = []
    segments.insert(0, "root")
    return ":".join(segments)


def eventstream(f):
    def new_func(*args, **kwargs):
        def generate():
            for event in chain(f(*args, **kwargs), (None,)):
                yield ("data: %s\n\n" % json.dumps(event)).encode()

        return Response(
            generate(), mimetype="text/event-stream", direct_passthrough=True
        )

    return update_wrapper(new_func, f)



from flask import Blueprint
from flask import render_template


bp = Blueprint("dash", __name__, url_prefix="/admin")


@bp.route("/")
@bp.route("/<view>", endpoint="app")
def app_view(**kwargs):
    """Render the React admin GUI app."""
    return render_template("dash.html")



# Tests for lektor.db.Tree and related classes.
import re
import shutil
from pathlib import Path

import pytest

from lektor.constants import PRIMARY_ALT
from lektor.db import Tree
from lektor.project import Project


@pytest.fixture(scope="session")
def no_alt_pad(tmp_path_factory):
    no_alt_project = tmp_path_factory.mktemp("no-alts") / "demo-project"
    demo_project = Path(__file__).parent / "demo-project"
    shutil.copytree(demo_project, no_alt_project)

    project_file = no_alt_project / "Website.lektorproject"
    alt_section_re = r"(?ms)^\[alternatives\.\w+\].*?(?=^\[)"
    project_file.write_text(re.sub(alt_section_re, "", project_file.read_text()))

    dirs = [no_alt_project]
    while dirs:
        for child in dirs.pop().iterdir():
            if child.is_dir():
                dirs.append(child)
            elif child.match("contents+*.lr"):
                child.unlink()

    project = Project.from_path(no_alt_project)
    return project.make_env().new_pad()


@pytest.fixture(params=[False, True])
def disable_alternatives(request):
    return request.param


@pytest.fixture
def tree(pad, no_alt_pad, disable_alternatives):
    if disable_alternatives:
        pad = no_alt_pad
    return Tree(pad)


@pytest.mark.parametrize(
    "path, expect",
    [
        ("/", ""),
        ("/blog", "blog"),
        ("/this/is/missing", "missing"),
    ],
)
def test_tree_item_id(tree, path, expect):
    item = tree.get(path)
    assert item.path == path
    assert item.id == expect


@pytest.mark.parametrize(
    "path, expect",
    [
        ("/", True),
        ("/test.jpg", True),
        ("/missing", False),
    ],
)
def test_tree_item_exists(tree, path, expect):
    item = tree.get(path)
    assert item.exists is expect
    if not item.exists:
        assert item._datamodel is None


@pytest.mark.parametrize(
    "path, expect",
    [
        ("/", False),
        ("/test.jpg", True),
        ("/missing", False),
    ],
)
def test_tree_item_can_be_deleted(tree, path, expect):
    item = tree.get(path)
    assert item.can_be_deleted == expect


@pytest.mark.parametrize(
    "path, expect",
    [
        ("/", True),
        ("/test.jpg", True),
        ("/missing", True),
        ("/extra/container", False),
    ],
)
def test_tree_item_is_visible(tree, path, expect):
    item = tree.get(path)
    assert item.is_visible == expect


@pytest.mark.parametrize(
    "path, is_attachment, attachment_type",
    [
        ("/", False, None),
        ("/test.jpg", True, "image"),
        ("/missing", False, None),
    ],
)
def test_tree_item_is_attachment(tree, path, is_attachment, attachment_type):
    item = tree.get(path)
    assert item.is_attachment is is_attachment
    assert item.attachment_type == attachment_type


@pytest.mark.parametrize(
    "path, expect",
    [
        ("/", True),
        ("/test.jpg", False),
        ("/missing", False),
    ],
)
def test_tree_item_can_have_children(tree, path, expect):
    item = tree.get(path)
    assert item.can_have_children == expect
    # NB: currently in the demo-project, can_have_children ==
    # can_have_attachments
    assert item.can_have_attachments == expect


@pytest.mark.parametrize(
    "disable_alternatives, path, alt, expect",
    [
        (False, "/", PRIMARY_ALT, {"en": "Welcome"}),
        (False, "/test.jpg", PRIMARY_ALT, {"en": "test.jpg"}),
        (False, "/missing/gone", PRIMARY_ALT, {"en": "Gone"}),
        (False, "/projects/bagpipe", PRIMARY_ALT, {"en": "Bagpipe"}),
        (False, "/projects/bagpipe", "de", {"en": "Dudelsack"}),
        (True, "/", PRIMARY_ALT, {"en": "Welcome"}),
        (True, "/test.jpg", PRIMARY_ALT, {"en": "test.jpg"}),
        (True, "/missing/gone", PRIMARY_ALT, {"en": "Gone"}),
        (True, "/projects/bagpipe", PRIMARY_ALT, {"en": "Bagpipe"}),
    ],
)
def test_tree_item_get_record_label_i18n(tree, path, alt, expect):
    item = tree.get(path)
    assert item.get_record_label_i18n(alt) == expect


@pytest.mark.parametrize(
    "path, expect",
    [
        ("/", None),
        ("/blog", "blog-post"),
    ],
)
def test_tree_implied_child_datamodel(tree, path, expect):
    item = tree.get(path)
    assert item.implied_child_datamodel == expect


@pytest.mark.parametrize(
    "path, parent_path",
    [
        ("/", None),
        ("/test.jpg", "/"),
        ("/missing/gone", "/missing"),
        ("/projects/bagpipe", "/projects"),
    ],
)
def test_get_parent(tree, path, parent_path):
    item = tree.get(path)
    parent = item.get_parent()
    assert (parent and parent.path) == parent_path


@pytest.mark.parametrize(
    "path, name",
    [
        ("/", "a"),
        ("/missing/gone", "foo"),
    ],
)
def test_get(tree, path, name):
    item = tree.get(path)
    assert item.get(name).id == name


@pytest.mark.parametrize(
    "disable_alternatives, path, offset, limit, order_by, expect",
    [
        (False, "/extra/container", 0, None, None, ["a", "hello.txt"]),
        (False, "/extra/container", 1, None, None, ["hello.txt"]),
        (False, "/extra/container", 0, 1, None, ["a"]),
        (False, "/projects", 0, 2, ("seq",), ["attachment.txt", "zaun"]),
        (False, "/projects", 1, 1, ("seq",), ["zaun"]),
        (False, "/projects", 0, 2, None, ["attachment.txt", "bagpipe"]),
        (False, "/projects", 8, None, None, ["slave", "wolf", "zaun"]),
        (True, "/extra/container", 0, None, None, ["a", "hello.txt"]),
        (True, "/extra/container", 1, None, None, ["hello.txt"]),
        (True, "/extra/container", 0, 1, None, ["a"]),
        (True, "/projects", 0, 2, ("seq",), ["attachment.txt", "wolf"]),
        (True, "/projects", 1, 1, ("seq",), ["wolf"]),
        (True, "/projects", 0, 2, None, ["attachment.txt", "bagpipe"]),
        (True, "/projects", 8, None, None, ["slave", "wolf"]),
    ],
)
def test_tree_item_get_children(tree, path, offset, limit, order_by, expect):
    item = tree.get(path)
    children = item.get_children(offset, limit, order_by=order_by)
    assert list(child.id for child in children) == expect


@pytest.mark.parametrize(
    "disable_alternatives, path, order_by, expect",
    [
        (False, "/", None, ["blog", "extra", "projects"]),
        (False, "/blog", ("-pub_date",), ["post2", "post1", "dummy.xml"]),
        (
            False,
            "/projects",
            ("seq",),
            [
                "zaun",
                "wolf",
                "slave",
                "secret",
                "postage",
                "oven",
                "master",
                "filtered",
                "bagpipe",
                "coffee",
            ],
        ),
        (
            False,
            "/projects",
            None,
            [
                "bagpipe",
                "coffee",
                "filtered",
                "master",
                "oven",
                "postage",
                "secret",
                "slave",
                "wolf",
                "zaun",
            ],
        ),
        (True, "/", None, ["blog", "extra", "projects"]),
        (True, "/blog", ("-pub_date",), ["post2", "post1", "dummy.xml"]),
    ],
)
def test_tree_item_iter_subpages(tree, path, order_by, expect):
    item = tree.get(path)
    assert list(child.id for child in item.iter_subpages(order_by)) == expect


@pytest.mark.parametrize(
    "path, expect",
    [
        (
            "/",
            [
                "hello.txt",
                "test-progressive.jpg",
                "test-sof-last.jpg",
                "test.jpg",
                "test.mp4",
            ],
        ),
    ],
)
def test_tree_item_iter_attachments(tree, path, expect):
    item = tree.get(path)
    assert list(child.id for child in item.iter_attachments()) == expect


@pytest.mark.parametrize(
    "path, order_by, expect",
    [
        ("/", ("title",), [("Welcome", False)]),
        ("/extra", ("title", "+_id"), [("Hello", False), ("extra", False)]),
        ("/missing", ("title", "-_id"), [(None, False), ("missing", True)]),
    ],
)
def test_tree_item_get_order_by(tree, path, order_by, expect):
    item = tree.get(path)
    sort_key = item.get_sort_key(order_by)
    assert [(_.value, _.reverse) for _ in sort_key] == expect


@pytest.mark.parametrize(
    "path, expect",
    [
        ("/", "<TreeItem '/'>"),
        ("/hello.txt", "<TreeItem '/hello.txt' attachment>"),
    ],
)
def test_tree_item_repr(tree, path, expect):
    item = tree.get(path)
    assert repr(item) == expect


@pytest.mark.parametrize(
    "disable_alternatives, alts",
    [(False, {PRIMARY_ALT, "en", "de"}), (True, {PRIMARY_ALT})],
)
def test_tree_item_alts(tree, alts):
    item = tree.get("/")
    assert set(item.alts) == alts


@pytest.mark.parametrize(
    "disable_alternatives, path, alt, expect",
    [
        (False, "/", PRIMARY_ALT, True),
        (False, "/", "en", False),
        (False, "/", "de", False),
        (False, "/hello.txt", PRIMARY_ALT, False),
        (False, "/hello.txt", "en", False),
        (False, "/hello.txt", "de", False),
        (False, "/missing", PRIMARY_ALT, False),
        (False, "/missing", "en", False),
        (False, "/missing", "de", False),
        (False, "/projects/zaun", PRIMARY_ALT, False),
        (False, "/projects/zaun", "en", False),
        (False, "/projects/zaun", "de", True),
        (True, "/", PRIMARY_ALT, True),
        (True, "/hello.txt", PRIMARY_ALT, False),
        (True, "/missing", PRIMARY_ALT, False),
        (True, "/projects/zaun", PRIMARY_ALT, False),
    ],
)
def test_alt_exists(tree, path, alt, expect):
    item = tree.get(path)
    assert item.alts[alt].exists == expect


@pytest.mark.parametrize(
    "disable_alternatives, alt, expect",
    [
        (False, PRIMARY_ALT, False),
        (False, "en", True),
        (False, "de", False),
        (True, PRIMARY_ALT, True),
    ],
)
def test_alt_is_primary_overlay(tree, alt, expect):
    item = tree.get("/")
    assert item.alts[alt].is_primary_overlay == expect


@pytest.mark.parametrize(
    "disable_alternatives, alt, expect",
    [
        (False, PRIMARY_ALT, {"en": "English", "de": "Englisch"}),
        (False, "en", {"en": "English", "de": "Englisch"}),
        (False, "de", {"en": "German", "de": "Deutsch"}),
    ],
)
def test_alt_name_i18n(tree, alt, expect):
    item = tree.get("/")
    assert item.alts[alt].name_i18n == expect


@pytest.mark.parametrize(
    "disable_alternatives, path, alt, expect",
    [
        (False, "/", PRIMARY_ALT, "<Alt '_primary'*>"),
        (False, "/", "en", "<Alt 'en'>"),
        (False, "/", "de", "<Alt 'de'>"),
        (False, "/projects/zaun", "de", "<Alt 'de'*>"),
        (True, "/", PRIMARY_ALT, "<Alt '_primary'*>"),
        (True, "/projects/zaun", PRIMARY_ALT, "<Alt '_primary'>"),
    ],
)
def test_alt_repr(tree, path, alt, expect):
    item = tree.get(path)
    assert repr(item.alts[alt]) == expect



import datetime

from babel.dates import get_timezone
from markupsafe import escape
from markupsafe import Markup

from lektor.context import Context
from lektor.datamodel import Field
from lektor.types.base import BadValue
from lektor.types.base import Undefined
from lektor.types.formats import MarkdownDescriptor


class DummySource:
    url_path = "/"


def make_field(env, type, **options):
    return Field(env, "demo", type=env.types[type], options=options)


def test_markdown(env, pad):
    field = make_field(env, "markdown")

    source = DummySource()

    with Context(pad=pad):
        rv = field.deserialize_value("Hello **World**!", pad=pad)
        assert isinstance(rv, MarkdownDescriptor)
        rv = rv.__get__(source)
        assert rv
        assert rv.source == "Hello **World**!"
        assert escape(rv) == Markup("<p>Hello <strong>World</strong>!</p>\n")
        assert rv.meta == {}

        for val in "", None:
            rv = field.deserialize_value(val, pad=pad)
            assert isinstance(rv, MarkdownDescriptor)
            rv = rv.__get__(source)
            assert not rv
            assert rv.source == ""
            assert escape(rv) == Markup("")
            assert rv.meta == {}


def test_markdown_links(env, pad):
    field = make_field(env, "markdown")
    source = DummySource()

    def md(s):
        rv = field.deserialize_value(s, pad=pad)
        assert isinstance(rv, MarkdownDescriptor)
        return str(rv.__get__(source)).strip()

    with Context(pad=pad):
        assert md("[foo](http://example.com/)") == (
            '<p><a href="http://example.com/">foo</a></p>'
        )
        assert md("[foo](javascript:foo)") == (
            '<p><a href="javascript:foo">foo</a></p>'
        )

        img = (
            "iVBORw0KGgoAAAANSUhEUgAAAAUAAAAFCAYAAACNbyblAAAAHElEQVQI12P4"
            "//8/w38GIAXDIBKE0DHxgljNBAAO9TXL0Y4OHwAAAABJRU5ErkJggg=="
        )
        assert (
            md("![test](data:image/png;base64,%s)" % img)
            == ('<p><img src="data:image/png;base64,%s" alt="test"></p>') % img
        )


def test_markdown_linking(pad, builder):
    blog_index = pad.get("/blog", page_num=1)

    prog, _ = builder.build(blog_index)
    with prog.artifacts[0].open("rb") as f:
        assert (
            b'Look at my <a href="2015/12/post1/hello.txt">' b"attachment</a>"
        ) in f.read()

    blog_post = pad.get("/blog/post1")

    prog, _ = builder.build(blog_post)
    with prog.artifacts[0].open("rb") as f:
        assert b'Look at my <a href="hello.txt">' b"attachment</a>" in f.read()


def test_markdown_images(pad, builder):
    blog_index = pad.get("/blog", page_num=1)

    prog, _ = builder.build(blog_index)
    with prog.artifacts[0].open("rb") as f:
        assert (
            b'This is an image <img src="2015/12/' b'post1/logo.png" alt="logo">.'
        ) in f.read()

    blog_post = pad.get("/blog/post1")

    prog, _ = builder.build(blog_post)
    with prog.artifacts[0].open("rb") as f:
        assert b'This is an image <img src="logo.png" alt="logo">.' in f.read()


def test_string(env, pad):
    field = make_field(env, "string")

    with Context(pad=pad):
        rv = field.deserialize_value("", pad=pad)
        assert rv == ""

        rv = field.deserialize_value(None, pad=pad)
        assert isinstance(rv, Undefined)

        rv = field.deserialize_value("foo\nbar", pad=pad)
        assert rv == "foo"

        rv = field.deserialize_value(" 123 ", pad=pad)
        assert rv == "123"


def test_text(env, pad):
    field = make_field(env, "text")

    with Context(pad=pad):
        rv = field.deserialize_value("", pad=pad)
        assert rv == ""

        rv = field.deserialize_value(None, pad=pad)
        assert isinstance(rv, Undefined)

        rv = field.deserialize_value("foo\nbar", pad=pad)
        assert rv == "foo\nbar"

        rv = field.deserialize_value(" 123 ", pad=pad)
        assert rv == " 123 "


def test_integer(env, pad):
    field = make_field(env, "integer")

    with Context(pad=pad):
        rv = field.deserialize_value("", pad=pad)
        assert isinstance(rv, BadValue)

        rv = field.deserialize_value(None, pad=pad)
        assert isinstance(rv, Undefined)

        rv = field.deserialize_value("42", pad=pad)
        assert rv == 42

        rv = field.deserialize_value(" 23 ", pad=pad)
        assert rv == 23


def test_float(env, pad):
    field = make_field(env, "float")

    with Context(pad=pad):
        rv = field.deserialize_value("", pad=pad)
        assert isinstance(rv, BadValue)

        rv = field.deserialize_value(None, pad=pad)
        assert isinstance(rv, Undefined)

        rv = field.deserialize_value("42", pad=pad)
        assert rv == 42.0

        rv = field.deserialize_value(" 23.0 ", pad=pad)
        assert rv == 23.0

        rv = field.deserialize_value("-23.5", pad=pad)
        assert rv == -23.5


def test_boolean(env, pad):
    field = make_field(env, "boolean")

    with Context(pad=pad):
        rv = field.deserialize_value("", pad=pad)
        assert isinstance(rv, BadValue)

        rv = field.deserialize_value(None, pad=pad)
        assert isinstance(rv, Undefined)

        for s in "true", "TRUE", "True", "1", "yes":
            rv = field.deserialize_value(s, pad=pad)
            assert rv is True

        for s in "false", "FALSE", "False", "0", "no":
            rv = field.deserialize_value(s, pad=pad)
            assert rv is False


def test_datetime(env, pad):
    field = make_field(env, "datetime")

    with Context(pad=pad):
        # default
        rv = field.deserialize_value("2016-04-30 01:02:03", pad=pad)
        assert isinstance(rv, datetime.datetime)
        assert rv.year == 2016
        assert rv.month == 4
        assert rv.day == 30
        assert rv.hour == 1
        assert rv.minute == 2
        assert rv.second == 3
        assert rv.tzinfo is None

        # It is not datetime, it is None
        rv = field.deserialize_value(None, pad=pad)
        assert isinstance(rv, Undefined)

        # It is not datetime, it is empty string
        rv = field.deserialize_value("", pad=pad)
        assert isinstance(rv, BadValue)

        # It is not datetime, it is date
        rv = field.deserialize_value("2016-04-30", pad=pad)
        assert isinstance(rv, BadValue)


def test_datetime_timezone_utc(env, pad):
    field = make_field(env, "datetime")
    with Context(pad=pad):
        # Known timezone name, UTC
        rv = field.deserialize_value("2016-04-30 01:02:03 UTC", pad=pad)
        assert isinstance(rv, datetime.datetime)
        assert rv.year == 2016
        assert rv.month == 4
        assert rv.day == 30
        assert rv.hour == 1
        assert rv.minute == 2
        assert rv.second == 3
        assert rv.tzinfo is get_timezone("UTC")


def test_datetime_timezone_est(env, pad):
    field = make_field(env, "datetime")
    with Context(pad=pad):
        # Known timezone name, EST
        rv = field.deserialize_value("2016-04-30 01:02:03 EST", pad=pad)
        assert isinstance(rv, datetime.datetime)
        assert rv.year == 2016
        assert rv.month == 4
        assert rv.day == 30
        assert rv.hour == 1
        assert rv.minute == 2
        assert rv.second == 3
        assert rv.tzinfo is get_timezone("EST")


def test_datetime_timezone_location(env, pad):
    field = make_field(env, "datetime")
    with Context(pad=pad):
        # Known location name, Asia/Seoul
        rv = field.deserialize_value("2016-04-30 01:02:03 Asia/Seoul", pad=pad)
        assert isinstance(rv, datetime.datetime)
        assert rv.year == 2016
        assert rv.month == 4
        assert rv.day == 30
        assert rv.hour == 1
        assert rv.minute == 2
        assert rv.second == 3
        tzinfos = get_timezone("Asia/Seoul")._tzinfos  # pylint: disable=no-member
        assert rv.tzinfo in tzinfos.values()


def test_datetime_timezone_kst(env, pad):
    field = make_field(env, "datetime")
    with Context(pad=pad):
        # KST - http://www.timeanddate.com/time/zones/kst
        rv = field.deserialize_value("2016-04-30 01:02:03 +0900", pad=pad)
        assert isinstance(rv, datetime.datetime)
        assert rv.year == 2016
        assert rv.month == 4
        assert rv.day == 30
        assert rv.hour == 1
        assert rv.minute == 2
        assert rv.second == 3
        assert rv.tzinfo._offset == datetime.timedelta(0, 9 * 60 * 60)


def test_datetime_timezone_acst(env, pad):
    field = make_field(env, "datetime")
    with Context(pad=pad):
        # ACST - http://www.timeanddate.com/time/zones/acst
        rv = field.deserialize_value("2016-04-30 01:02:03 +0930", pad=pad)
        assert isinstance(rv, datetime.datetime)
        assert rv.year == 2016
        assert rv.month == 4
        assert rv.day == 30
        assert rv.hour == 1
        assert rv.minute == 2
        assert rv.second == 3
        assert rv.tzinfo._offset == datetime.timedelta(0, (9 * 60 + 30) * 60)


def test_datetime_timezone_mst(env, pad):
    field = make_field(env, "datetime")
    with Context(pad=pad):
        # MST - http://www.timeanddate.com/time/zones/mst
        rv = field.deserialize_value("2016-04-30 01:02:03 -0700", pad=pad)
        assert isinstance(rv, datetime.datetime)
        assert rv.year == 2016
        assert rv.month == 4
        assert rv.day == 30
        assert rv.hour == 1
        assert rv.minute == 2
        assert rv.second == 3
        assert rv.tzinfo._offset == datetime.timedelta(0, -7 * 60 * 60)


def test_datetime_timezone_mart(env, pad):
    field = make_field(env, "datetime")
    with Context(pad=pad):
        # MART - http://www.timeanddate.com/time/zones/mart
        rv = field.deserialize_value("2016-04-30 01:02:03 -0930", pad=pad)
        assert isinstance(rv, datetime.datetime)
        assert rv.year == 2016
        assert rv.month == 4
        assert rv.day == 30
        assert rv.hour == 1
        assert rv.minute == 2
        assert rv.second == 3
        assert rv.tzinfo._offset == datetime.timedelta(0, -(9 * 60 + 30) * 60)


def test_datetime_timezone_name(env, pad):
    field = make_field(env, "datetime")
    with Context(pad=pad):
        # with timezone name (case 1)
        rv = field.deserialize_value("2016-04-30 01:02:03 KST +0900", pad=pad)
        assert isinstance(rv, datetime.datetime)
        assert rv.year == 2016
        assert rv.month == 4
        assert rv.day == 30
        assert rv.hour == 1
        assert rv.minute == 2
        assert rv.second == 3
        assert rv.tzinfo._offset == datetime.timedelta(0, 9 * 60 * 60)

        # with timezone name (case 2)
        rv = field.deserialize_value("2016-04-30 01:02:03 KST+0900", pad=pad)
        assert isinstance(rv, datetime.datetime)
        assert rv.year == 2016
        assert rv.month == 4
        assert rv.day == 30
        assert rv.hour == 1
        assert rv.minute == 2
        assert rv.second == 3
        assert rv.tzinfo._offset == datetime.timedelta(0, 9 * 60 * 60)



"""Test that all modules/packages in the lektor tree are importable in any order

Here we import each module by itself, one at a time, each in a new
python interpreter.

"""
import pkgutil
import sys
from subprocess import run

import pytest

import lektor


def iter_lektor_modules():
    for module in pkgutil.walk_packages(lektor.__path__, f"{lektor.__name__}."):
        yield module.name


@pytest.fixture(params=iter_lektor_modules())
def module(request):
    return request.param


def test_import(module):
    python = sys.executable
    assert run([python, "-c", f"import {module}"], check=False).returncode == 0



import flask
import pytest
from werkzeug.exceptions import HTTPException

from lektor.admin.webui import WebAdmin
from lektor.constants import PRIMARY_ALT


app = flask.Flask(__name__)


@pytest.fixture
def output_path(tmp_path):
    output_path = tmp_path / "webadmin"
    output_path.mkdir()
    return output_path


@pytest.fixture
def webadmin(env, output_path):
    return WebAdmin(env, output_path=output_path.__fspath__())


@pytest.fixture
def lektor_info(webadmin):
    return webadmin.lektor_info


@pytest.fixture
def resolve_artifact(webadmin):
    def resolve(url_path):
        with app.test_request_context(url_path):
            return webadmin.lektor_info.resolve_artifact(url_path)

    return resolve


@pytest.mark.parametrize(
    "url_path, expected",
    [
        (
            "/dir_with_index_html/",
            (
                "dir_with_index_html/index.html",
                "dir_with_index_html/index.html",
                None,
                PRIMARY_ALT,
            ),
        ),
        (
            "/dir_with_index_htm/",
            (
                "dir_with_index_htm/index.htm",
                "dir_with_index_htm/index.htm",
                None,
                PRIMARY_ALT,
            ),
        ),
        (
            "/empty",
            # Resolves to Directory, so has alt
            (None, "empty", None, PRIMARY_ALT),
        ),
        ("/missing", (None, "missing", None, None)),  # Does not resolve, no alt
        ("/extra/", ("extra/index.html", "extra/index.html", "/extra", "en")),
        (
            "/de/extra/long/path/",
            (
                "de/extra/long/path/index.html",
                "de/extra/long/path/index.html",
                "/extra/slash-slug",
                "de",
            ),
        ),
    ],
)
def test_resolve_artifact(resolve_artifact, url_path, expected, output_path):
    def massage_expected(expected):
        artifact_name, filename, record_path, alt = expected
        # NB: this converts filename to native (e.g. pathsep=='\\' on windows) path
        filename = output_path.joinpath(filename).__fspath__()
        return artifact_name, filename, record_path, alt

    assert resolve_artifact(url_path) == massage_expected(expected)


@pytest.mark.parametrize(
    "url_path, location",
    [
        ("/dir_with_index_html", "dir_with_index_html/"),
        ("/dir_with_index_htm", "dir_with_index_htm/"),
    ],
)
def test_resolve_artifact_redirects(resolve_artifact, url_path, location):
    with pytest.raises(HTTPException) as exc:
        resolve_artifact(url_path)
    # add_slash_redirect in werkzeug>=2.1.0 returns 308 - previously it returned 301
    assert exc.value.response.status_code in (301, 308)
    assert exc.value.response.headers["Location"] == location


def test_get_admin_does_something_useful(webadmin, mocker):
    # Test that GET /admin eventually gets to the admin JS app
    # See https://github.com/lektor/lektor/issues/1043
    render_template = mocker.patch(
        "lektor.admin.modules.dash.render_template",
        return_value="RENDERED",
    )
    with webadmin.test_client() as test_client:
        resp = test_client.get("/admin", follow_redirects=True)
    assert resp.status_code == 200
    assert resp.get_data(as_text=True) == render_template.return_value
    assert render_template.mock_calls == [mocker.call("dash.html")]



class Expression:
    def __init__(self, env, expr):
        self.env = env
        self.tmpl = env.jinja_env.from_string("{{ __result__(%s) }}" % expr)

    def evaluate(self, pad=None, this=None, values=None, alt=None):
        result = []

        def result_func(value):
            result.append(value)
            return ""

        values = self.env.make_default_tmpl_values(pad, this, values, alt)
        values["__result__"] = result_func
        self.tmpl.render(values)
        return result[0]


class FormatExpression:
    def __init__(self, env, expr):
        self.env = env
        self.tmpl = env.jinja_env.from_string(expr)

    def evaluate(self, pad=None, this=None, values=None, alt=None):
        values = self.env.make_default_tmpl_values(pad, this, values, alt)
        return self.tmpl.render(values)



import gc
import os
import sys
import warnings
import weakref
from shutil import which

import pytest

from lektor.publisher import Command
from lektor.publisher import publish


def test_Command_triggers_no_warnings():
    # This excercises the issue where publishing via rsync resulted
    # in ResourceWarnings about unclosed streams.

    with warnings.catch_warnings():
        warnings.simplefilter("error")

        # This is essentially how RsyncPublisher runs rsync.
        with Command([sys.executable, "-c", "print()"]) as client:
            for _ in client:
                pass

        # The ResourceWarnings regarding unclosed files we are checking for
        # are issued during finalization.  Without this extra effort,
        # finalization wouldn't happen until after the test completes.
        client_is_alive = weakref.ref(client)
        del client
        if client_is_alive():
            gc.collect()

    if client_is_alive():
        warnings.warn(
            "Unable to trigger garbage collection of Command instance, "
            "so unable to check for warnings issued during finalization."
        )


@pytest.mark.skipif(
    which("rsync") is None, reason="rsync is not available on this system"
)
@pytest.mark.parametrize("delete", ["yes", "no"])
def test_RsyncPublisher_integration(env, tmp_path, delete):
    # Integration test of local rsync deployment
    # Ensures that RsyncPublisher can successfully invoke rsync
    files = {"file.txt": "content\n"}
    output = tmp_path / "output"
    output.mkdir()
    for path, content in files.items():
        output.joinpath(path).write_text(content)

    target_path = tmp_path / "target"
    target_path.mkdir()
    target = f"rsync://{target_path.resolve()}?delete={delete}"

    event_iter = publish(env, target, output)
    for line in event_iter:
        print(line)

    target_files = {
        os.fspath(_.relative_to(target_path)): _.read_text()
        for _ in target_path.iterdir()
    }
    assert target_files == files



import pytest

from lektor.utils import cleanup_path


def test_cleanup_path():

    assert cleanup_path("/") == "/"
    assert cleanup_path("/foo") == "/foo"
    assert cleanup_path("/foo/") == "/foo"
    assert cleanup_path("/////foo/") == "/foo"
    assert cleanup_path("/////foo////") == "/foo"
    assert cleanup_path("/////foo/.///") == "/foo"
    assert cleanup_path("/////foo/..///") == "/foo"
    assert cleanup_path("/foo/./bar/") == "/foo/bar"
    assert cleanup_path("/foo/../bar/") == "/foo/bar"


def test_basic_url_to_with_alts(pad):

    wolf_en = pad.get("/projects/wolf", alt="en")
    slave_en = pad.get("/projects/slave", alt="en")
    wolf_de = pad.get("/projects/wolf", alt="de")
    slave_de = pad.get("/projects/slave", alt="de")

    assert wolf_en.url_to(slave_en) == "../slave/"
    assert wolf_de.url_to(slave_de) == "../sklave/"
    assert slave_en.url_to(slave_de) == "../../de/projects/sklave/"
    assert slave_de.url_to(slave_en) == "../../../projects/slave/"


@pytest.mark.parametrize(
    "path, alt, absolute, external, base_url, expected",
    [
        ("/projects/slave", "", None, None, None, "../slave/"),
        ("/projects/slave", "de", None, None, None, "../../de/projects/sklave/"),
        ("/projects/slave", "de", True, None, None, "/de/projects/sklave/"),
        ("/projects/slave", "de", True, True, None, "/de/projects/sklave/"),
        ("/projects/slave", "de", True, True, "/content/", "/de/projects/sklave/"),
        (
            "/projects/slave",
            "de",
            None,
            True,
            None,
            "/projects/slave1/de/projects/sklave/",
        ),
        (
            "/projects/slave",
            "de",
            None,
            True,
            "/content/",
            "/projects/slave1/de/projects/sklave/",
        ),
        ("/projects/slave", "de", None, None, "/content/", "../de/projects/sklave/"),
        ("", "de", None, None, None, "../../de/projects/wolf/"),
        ("!", "de", None, None, None, "./"),
        ("!/projects/slave", "de", None, None, None, "../slave"),
        ("!/projects/slave", None, None, None, None, "../slave"),
        ("!", None, None, None, None, "./"),
        ("", None, None, None, None, "./"),
        ("/projects/slave", None, True, None, None, "/projects/slave/"),
        ("/projects/slave", None, True, True, None, "/projects/slave/"),
        ("/projects/slave", None, True, True, "/content/", "/projects/slave/"),
        ("/projects/slave", None, True, None, "/content/", "/projects/slave/"),
        ("/projects/slave", None, None, True, None, "/projects/slave1/projects/slave/"),
        (
            "/projects/slave",
            None,
            None,
            True,
            "/content/",
            "/projects/slave1/projects/slave/",
        ),
        ("/projects/slave", None, None, None, "/content/", "../projects/slave/"),
        ("/projects/slave", None, None, None, None, "../slave/"),
    ],
)
def test_url_to_all_params(pad, path, alt, absolute, external, base_url, expected):

    if external and not absolute:
        pad.db.config.base_url = "/projects/slave1/"

    wolf_en = pad.get("/projects/wolf")

    assert wolf_en.url_to(path, alt, absolute, external, base_url) == expected


@pytest.mark.parametrize(
    "path, alt, absolute, external, base_url",
    [
        ("/projects/slave", "de", None, True, None),
        ("/projects/slave", "de", None, True, "/content/"),
        ("/projects/slave", None, None, True, None),
        ("/projects/slave", None, None, True, "/content/"),
    ],
)
def test_url_to_all_params_error_cases(pad, path, alt, absolute, external, base_url):

    wolf_en = pad.get("/projects/wolf")

    with pytest.raises(RuntimeError):
        wolf_en.url_to(path, alt, absolute, external, base_url)


def test_url_parse_virtual_path(pad):
    projects_de = pad.get("/projects@2", alt="de")
    paginated_de = pad.get("/extra/paginated@2", alt="de")
    paginated = pad.get("/extra/paginated@2")
    assert projects_de.url_path == "/de/projects/page/2/"
    assert paginated_de.url_path == "/de/extra/paginiert/page/2/"
    assert paginated.url_path == "/extra/paginated/page/2/"


def test_file_record_url(pad):
    record = pad.get("/extra/file.ext")
    assert record.url_path == "/extra/file.ext"


def test_url_with_slash_slug(pad):
    record = pad.get("/extra/slash-slug")
    assert record.url_path == "/extra/long/path/"



import os
from datetime import timedelta

import pytest

from lektor.utils import locate_executable
from lektor.videotools import Dimensions
from lektor.videotools import get_ffmpeg_quality
from lektor.videotools import get_timecode


has_ffmpeg = locate_executable("ffmpeg")
has_ffprobe = locate_executable("ffprobe")
require_ffmpeg = pytest.mark.skipif(
    not has_ffmpeg or not has_ffprobe, reason="requires ffmpeg and ffprobe in path"
)


@pytest.mark.parametrize(
    "td, tc",
    [
        (timedelta(), "00:00:00"),
        (timedelta(seconds=1), "00:00:01"),
        (timedelta(seconds=1.5), "00:00:01.5"),
        (timedelta(hours=8, minutes=15), "08:15:00"),
        (timedelta(hours=36, minutes=35, seconds=34.25), "36:35:34.25"),
    ],
)
def test_get_timecode(td, tc):
    assert get_timecode(td) == tc


@pytest.mark.parametrize(
    "in_quality, out_quality",
    [
        (0, 31),
        (100, 2),
    ],
)
def test_get_ffmpeg_quality(in_quality, out_quality):
    assert get_ffmpeg_quality(in_quality) == out_quality


@pytest.mark.parametrize(
    "in_quality",
    [
        -1,
        101,
    ],
)
def test_get_ffmpeg_quality_bad_range(in_quality):
    with pytest.raises(ValueError):
        get_ffmpeg_quality(in_quality)


@pytest.mark.parametrize(
    "dim, ar",
    [
        (Dimensions(100, 100), 1),
        (Dimensions(320, 160), 2),
        (Dimensions(160, 320), 0.5),
    ],
)
def test_dimensions_aspect_ratio(dim, ar):
    assert dim.aspect_ratio == ar


@pytest.mark.parametrize(
    "source, target, upscale, rescale",
    [
        ((100, 100), (50, None), None, (50, 50)),
        ((100, 50), (50, None), None, (50, 25)),
        ((100, 50), (None, 25), None, (50, 25)),
    ],
)
def test_dimensions_fit_within(source, target, upscale, rescale):
    # We don't take a crop parameter here since it will always be the same as
    # rescale
    source = Dimensions(*source)
    target_width, target_height = target
    rescale = Dimensions(*rescale)

    new_dims = source.fit_within(target_width, target_height, upscale)
    assert rescale == new_dims.rescale
    assert rescale == new_dims.crop


@pytest.mark.parametrize(
    "source, target, upscale, rescale, crop",
    [
        ((100, 50), (25, 25), None, (50, 25), (25, 25)),
        ((100, 50), (100, 100), None, (200, 100), (100, 100)),
        ((100, 50), (100, 100), False, (100, 50), (100, 50)),
    ],
)
def test_dimensions_cover(source, target, upscale, rescale, crop):
    source = Dimensions(*source)
    target_width, target_height = target
    rescale = Dimensions(*rescale)
    crop = Dimensions(*crop)

    new_dims = source.cover(target_width, target_height, upscale)
    assert rescale == new_dims.rescale
    assert crop == new_dims.crop


@pytest.mark.parametrize(
    "source, target, upscale, rescale",
    [
        ((100, 50), (25, 25), None, (25, 25)),
        ((100, 50), (100, 100), None, (100, 100)),
        ((100, 50), (100, 100), False, (100, 50)),
    ],
)
def test_dimensions_stretch(source, target, upscale, rescale):
    # We don't take a crop parameter here since it will always be the same as
    # rescale
    source = Dimensions(*source)
    target_width, target_height = target
    rescale = Dimensions(*rescale)

    new_dims = source.stretch(target_width, target_height, upscale)
    assert rescale == new_dims.rescale
    assert rescale == new_dims.crop


def test_dimensions_invalid_resize():
    dim = Dimensions(640, 480)
    with pytest.raises(ValueError):
        dim.resize(mode="invalid")


@require_ffmpeg
def test_metadata(pad):
    video = pad.root.attachments.videos.get("test.mp4")
    assert video is not None

    assert video.width == 320
    assert video.height == 180
    assert video.duration == timedelta(seconds=3)


@require_ffmpeg
def test_thumbnail_height(builder):
    builder.build_all()
    with open(
        os.path.join(builder.destination_path, "index.html"), encoding="utf-8"
    ) as f:
        html = f.read()

    # The first thumbnail has the same dimensions as the source video,
    # seeked to 1.5 seconds in
    assert '<img src="test@t00-00-01-5.jpg" width="320" height="180">' in html

    # Thumbnail is half the original width, so its computed height is half.
    assert '<img src="test@t00-00-00_160.jpg" width="160" height="90">' in html

    # There should also be a square thumbnail
    assert (
        '<img src="test@t00-00-02_160x160_crop.jpg" width="160" height="160">' in html
    )



import re

from jinja2 import is_undefined
from jinja2 import TemplateNotFound
from markupsafe import Markup

from lektor.constants import PRIMARY_ALT
from lektor.context import get_ctx
from lektor.metaformat import tokenize
from lektor.types.base import Type


_block_re = re.compile(r"^####\s*([^#]*?)\s*####\s*$")
_line_unescape_re = re.compile(r"^#####(.*?)#####(\s*)$")


def discover_relevant_flowblock_models(flow, pad, record, alt):
    """Returns a dictionary of all relevant flow blocks.  If no list of
    flow block names is provided all flow blocks are returned.  Otherwise
    only flow blocks that are in the list or are children of flowblocks
    in the list are returned.
    """
    flow_blocks = flow.flow_blocks

    all_blocks = pad.db.flowblocks
    if flow_blocks is None:
        return dict((k, v.to_json(pad, record, alt)) for k, v in all_blocks.items())

    wanted_blocks = set()
    to_process = flow_blocks[:]

    while to_process:
        block_name = to_process.pop()
        flowblock = all_blocks.get(block_name)
        if block_name in wanted_blocks or flowblock is None:
            continue
        wanted_blocks.add(block_name)
        for field in flowblock.fields:
            if isinstance(field.type, FlowType):
                if field.type.flow_blocks is None:
                    raise RuntimeError(
                        "Nested flow-blocks require explicit "
                        "list of involved blocks."
                    )
                to_process.extend(field.type.flow_blocks)

    rv = {}
    for block_name in wanted_blocks:
        rv[block_name] = all_blocks[block_name].to_json(pad, record, alt)

    return rv


class BadFlowBlock(Exception):
    pass


class FlowBlock:
    """Represents a flowblock for the template."""

    def __init__(self, data, pad, record):
        self._data = data
        self._bound_data = {}
        self.pad = pad
        self.record = record

    @property
    def flowblockmodel(self):
        """The flowblock model that created this flow block."""
        return self.pad.db.flowblocks[self._data["_flowblock"]]

    def __contains__(self, name):
        return name in self._data and not is_undefined(self._data[name])

    def __getitem__(self, name):
        # If any data of a flowblock is accessed, we record that we need
        # this dependency.
        ctx = get_ctx()
        if ctx is not None:
            ctx.record_dependency(self.flowblockmodel.filename)

        rv = self._bound_data.get(name, Ellipsis)
        if rv is not Ellipsis:
            return rv
        rv = self._data[name]
        if hasattr(rv, "__get__"):
            rv = rv.__get__(self.record)
            self._bound_data[name] = rv
        return rv

    def __html__(self):
        ctx = get_ctx()

        # If we're in a nested render, we disable the rendering here or we
        # risk a recursion error.
        if ctx is None or self in ctx.flow_block_render_stack:
            return Markup.escape(repr(self))

        ctx.flow_block_render_stack.append(self)
        try:
            try:
                return self.pad.db.env.render_template(
                    [
                        "blocks/%s.html" % self._data["_flowblock"],
                        "blocks/default.html",
                    ],
                    pad=self.pad,
                    this=self,
                    alt=self.record.alt,
                    values={"record": self.record},
                )
            except TemplateNotFound:
                return Markup("[could not find snippet template]")
        finally:
            ctx.flow_block_render_stack.pop()

    def __repr__(self):
        return "<%s %r>" % (
            self.__class__.__name__,
            self["_flowblock"],
        )


class Flow:
    def __init__(self, blocks, record):
        self.blocks = blocks
        self.record = record

    def __html__(self):
        return Markup("\n\n".join(x.__html__() for x in self.blocks))

    def __bool__(self):
        return bool(self.blocks)

    __nonzero__ = __bool__

    def __repr__(self):
        return "<%s %r>" % (
            self.__class__.__name__,
            self.blocks,
        )


class FlowDescriptor:
    def __init__(self, blocks, pad):
        self._blocks = blocks
        self._pad = pad

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        return Flow([FlowBlock(data, self._pad, obj) for data in self._blocks], obj)


def process_flowblock_data(raw_value):
    lineiter = iter(raw_value.splitlines(True))
    block = None
    buf = []
    blocks = []

    for line in lineiter:
        # Until we found the first block, we ignore leading whitespace.
        if block is None and not line.strip():
            continue

        # Find a new block start
        block_start = _block_re.match(line)
        if block_start is None:
            if block is None:
                raise BadFlowBlock("Did not find beginning of flow block")
        else:
            if block is not None:
                blocks.append((block, buf))
                buf = []
            block = block_start.group(1)
            continue
        buf.append(_line_unescape_re.sub("####\\1####\\2", line))

    if block is not None:
        blocks.append((block, buf))

    return blocks


class FlowType(Type):
    widget = "flow"

    def __init__(self, env, options):
        Type.__init__(self, env, options)
        self.flow_blocks = [
            x.strip() for x in options.get("flow_blocks", "").split(",") if x.strip()
        ] or None

    def value_from_raw(self, raw):
        if raw.value is None:
            return raw.missing_value("Missing flow")
        if raw.pad is None:
            return raw.missing_value(
                "Flow value was technically present "
                "but used in a place where it cannot "
                "be used."
            )

        db = raw.pad.db
        rv = []

        try:
            for block, block_lines in process_flowblock_data(raw.value):
                # Unknown flow blocks are skipped for the moment
                if self.flow_blocks is not None and block not in self.flow_blocks:
                    continue
                flowblock = db.flowblocks.get(block)
                if flowblock is None:
                    continue

                d = {}
                for key, lines in tokenize(block_lines):
                    d[key] = "".join(lines)
                rv.append(flowblock.process_raw_data(d, pad=raw.pad))
        except BadFlowBlock as e:
            return raw.bad_value(str(e))

        return FlowDescriptor(rv, raw.pad)

    def to_json(self, pad, record=None, alt=PRIMARY_ALT):
        rv = Type.to_json(self, pad, record, alt)

        rv["flowblocks"] = discover_relevant_flowblock_models(self, pad, record, alt)

        block_order = self.flow_blocks
        if block_order is None:
            block_order = [
                k
                for k, v in sorted(pad.db.flowblocks.items(), key=lambda x: x[1].order)
            ]
        rv["flowblock_order"] = block_order

        return rv



import importlib

import lektor.i18n


def test_loading_i18n_triggers_no_warnings(recwarn):
    importlib.reload(lektor.i18n)
    for warning in recwarn.list:
        print(warning)  # debugging: display warnings on stdout
    assert len(recwarn) == 0
